from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from rattr.analyser.file import FileAnalyser
from rattr.models.context import compile_root_context
from rattr.models.ir import FileIr
from rattr.models.symbol import (
    Builtin,
    Call,
    CallArguments,
    CallInterface,
    Class,
    Func,
    Name,
)

if TYPE_CHECKING:
    from collections.abc import Iterator

    from tests.shared import MakeRootContextFn, ParseFn, StateFn


@pytest.fixture(autouse=True)
def __set_current_file(state: StateFn) -> Iterator[None]:
    with state(current_file=Path(__file__)):
        yield


class TestFunctionAnalyser:
    def test_basic_function(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
    ):
        ast_ = parse(
            """
            def a_func(arg):
                arg.sets_me = "value"
                return arg.gets_me
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        a_func = Func(name="a_func", interface=CallInterface(args=("arg",)))

        expected = FileIr(
            context=make_root_context([a_func], include_root_symbols=True),
            file_ir={
                a_func: {
                    "gets": {Name("arg.gets_me", "arg")},
                    "sets": {Name("arg.sets_me", "arg")},
                    "dels": set(),
                    "calls": set(),
                }
            },
        )

        assert results == expected

    def test_multiple_functions(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
    ):
        ast_ = parse(
            """
            def a_func(arg):
                arg.sets_me = "value"
                return arg.gets_me

            def another_func(arg):
                arg.attr = "this function only sets"
                arg.attr_two = "see!"
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        a_func = Func(name="a_func", interface=CallInterface(args=("arg",)))
        another_func = Func(name="another_func", interface=CallInterface(args=("arg",)))

        expected = FileIr(
            context=make_root_context(
                [
                    a_func,
                    another_func,
                ],
                include_root_symbols=True,
            ),
            file_ir={
                a_func: {
                    "gets": {Name("arg.gets_me", "arg")},
                    "sets": {Name("arg.sets_me", "arg")},
                    "dels": set(),
                    "calls": set(),
                },
                another_func: {
                    "gets": set(),
                    "sets": {
                        Name("arg.attr", "arg"),
                        Name("arg.attr_two", "arg"),
                    },
                    "dels": set(),
                    "calls": set(),
                },
            },
        )

        assert results == expected

    def test_conditional(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
    ):
        ast_ = parse(
            """
            def a_func(arg):
                if debug:
                    return False
                return arg.attr == "target value"
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        a_func = Func(name="a_func", interface=CallInterface(args=("arg",)))

        expected = FileIr(
            context=make_root_context(
                [a_func],
                include_root_symbols=True,
            ),
            file_ir={
                a_func: {
                    "gets": {
                        Name("debug"),
                        Name("arg.attr", "arg"),
                    },
                    "sets": set(),
                    "dels": set(),
                    "calls": set(),
                },
            },
        )

        assert results == expected

    def test_nested_conditional(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
    ):
        ast_ = parse(
            """
            def a_func(arg):
                if c1:
                    if c2:
                        return arg.foo
                return arg.bar
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        a_func = Func(name="a_func", interface=CallInterface(args=("arg",)))

        expected = FileIr(
            context=make_root_context(
                [a_func],
                include_root_symbols=True,
            ),
            file_ir={
                a_func: {
                    "gets": {
                        Name("c1"),
                        Name("c2"),
                        Name("arg.foo", "arg"),
                        Name("arg.bar", "arg"),
                    },
                    "sets": set(),
                    "dels": set(),
                    "calls": set(),
                },
            },
        )

        assert results == expected

    def test_nested_function(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
    ):
        # NOTE Does not test function following, `arg` always named such
        ast_ = parse(
            """
            def a_func(arg):
                def inner(arg):
                    return arg.foo
                return inner(arg)
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        a_func = Func(name="a_func", interface=CallInterface(args=("arg",)))
        inner_symbol = Func(name="inner", interface=CallInterface(args=("arg",)))
        inner_symbol_call = Call(
            name="inner",
            args=CallArguments(args=("arg",), kwargs={}),
            target=inner_symbol,
        )

        expected = FileIr(
            context=make_root_context(
                [a_func],
                include_root_symbols=True,
            ),
            file_ir={
                a_func: {
                    "gets": {
                        Name("arg"),
                        Name("arg.foo", "arg"),
                    },
                    "sets": set(),
                    "dels": set(),
                    "calls": {inner_symbol_call},
                },
            },
        )

        assert results == expected

    # def test_comprehensions(self, parse: ParseFn):
    #     ast_ = parse("""
    #         def list_comp(arg):
    #             return [a.prop for a in arg.iter]
    #     """)
    #     results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

    #     expected = {
    #         Func("list_comp", ["arg"], None, None): {
    #             "gets": {
    #                 # ...
    #             },
    #             "sets": set()
    #             "dels": set(),
    #             "calls": set(),
    #         },
    #     }

    #     assert results == expected

    def test_getattr_simple(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
    ):
        ast_ = parse(
            """
            def a_func(arg):
                return getattr(arg, "attr")
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        a_func = Func(name="a_func", interface=CallInterface(args=("arg",)))
        expected = FileIr(
            context=make_root_context(
                [a_func],
                include_root_symbols=True,
            ),
            file_ir={
                a_func: {
                    "gets": {Name("arg"), Name("arg.attr", "arg")},
                    "sets": set(),
                    "dels": set(),
                    "calls": set(),
                },
            },
        )

        assert results == expected

    def test_getattr_nested(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
    ):
        ast_ = parse(
            """
            def a_func(arg):
                return getattr(getattr(arg, "inner"), "outer")
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        a_func = Func(name="a_func", interface=CallInterface(args=("arg",)))
        expected = FileIr(
            context=make_root_context(
                [a_func],
                include_root_symbols=True,
            ),
            file_ir={
                a_func: {
                    "gets": {
                        Name("arg"),
                        Name("arg.inner", "arg"),
                        Name("arg.inner.outer", "arg"),
                    },
                    "sets": set(),
                    "dels": set(),
                    "calls": set(),
                },
            },
        )

        assert results == expected

    def test_getattr_nested_complex(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
    ):
        ast_ = parse(
            """
            def a_func(arg):
                return getattr(getattr(arg.b[0], "inner"), "outer")
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        a_func = Func(name="a_func", interface=CallInterface(args=("arg",)))
        expected = FileIr(
            context=make_root_context(
                [a_func],
                include_root_symbols=True,
            ),
            file_ir={
                a_func: {
                    "gets": {
                        Name("arg"),
                        Name("arg.b[]", "arg"),
                        Name("arg.b[].inner", "arg"),
                        Name("arg.b[].inner.outer", "arg"),
                    },
                    "sets": set(),
                    "dels": set(),
                    "calls": set(),
                },
            },
        )

        assert results == expected

    def test_hasattr_simple(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
    ):
        ast_ = parse(
            """
            def a_func(arg):
                return hasattr(arg, "attr")
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        a_func = Func(name="a_func", interface=CallInterface(args=("arg",)))
        expected = FileIr(
            context=make_root_context(
                [a_func],
                include_root_symbols=True,
            ),
            file_ir={
                a_func: {
                    "gets": {Name("arg"), Name("arg.attr", "arg")},
                    "sets": set(),
                    "dels": set(),
                    "calls": set(),
                },
            },
        )

        assert results == expected

    def test_hasattr_nested(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
    ):
        ast_ = parse(
            """
            def a_func(arg):
                return hasattr(hasattr(arg, "inner"), "outer")
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        a_func = Func(name="a_func", interface=CallInterface(args=("arg",)))
        expected = FileIr(
            context=make_root_context(
                [a_func],
                include_root_symbols=True,
            ),
            file_ir={
                a_func: {
                    "gets": {
                        Name("arg"),
                        Name("arg.inner", "arg"),
                        Name("arg.inner.outer", "arg"),
                    },
                    "sets": set(),
                    "dels": set(),
                    "calls": set(),
                },
            },
        )

        assert results == expected

    def test_hasattr_nested_complex(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
    ):
        ast_ = parse(
            """
            def a_func(arg):
                return hasattr(hasattr(arg.b[0], "inner"), "outer")
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        a_func = Func(name="a_func", interface=CallInterface(args=("arg",)))
        expected = FileIr(
            context=make_root_context(
                [a_func],
                include_root_symbols=True,
            ),
            file_ir={
                a_func: {
                    "gets": {
                        Name("arg"),
                        Name("arg.b[]", "arg"),
                        Name("arg.b[].inner", "arg"),
                        Name("arg.b[].inner.outer", "arg"),
                    },
                    "sets": set(),
                    "dels": set(),
                    "calls": set(),
                },
            },
        )

        assert results == expected

    def test_setattr_simple(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
    ):
        ast_ = parse(
            """
            def a_func(arg):
                setattr(arg, "attr", "value")
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        a_func = Func(name="a_func", interface=CallInterface(args=("arg",)))
        expected = FileIr(
            context=make_root_context(
                [a_func],
                include_root_symbols=True,
            ),
            file_ir={
                a_func: {
                    "gets": {Name("arg")},
                    "sets": {Name("arg.attr", "arg")},
                    "dels": set(),
                    "calls": set(),
                },
            },
        )

        assert results == expected

    def test_setattr_complex(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
    ):
        ast_ = parse(
            """
            def a_func(arg):
                return setattr(arg.b[0], "attr", "value")
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        a_func = Func(name="a_func", interface=CallInterface(args=("arg",)))
        expected = FileIr(
            context=make_root_context(
                [a_func],
                include_root_symbols=True,
            ),
            file_ir={
                a_func: {
                    "gets": {
                        Name("arg"),
                        Name("arg.b[]", "arg"),
                    },
                    "sets": {Name("arg.b[].attr", "arg")},
                    "dels": set(),
                    "calls": set(),
                },
            },
        )

        assert results == expected

    def test_delattr_simple(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
    ):
        ast_ = parse(
            """
            def a_func(arg):
                delattr(arg, "attr")
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        a_func = Func(name="a_func", interface=CallInterface(args=("arg",)))
        expected = FileIr(
            context=make_root_context(
                [a_func],
                include_root_symbols=True,
            ),
            file_ir={
                a_func: {
                    "gets": {Name("arg")},
                    "sets": set(),
                    "dels": {Name("arg.attr", "arg")},
                    "calls": set(),
                },
            },
        )

        assert results == expected

    def test_delattr_complex(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
    ):
        ast_ = parse(
            """
            def a_func(arg):
                return delattr(arg.b[0], "attr")
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        a_func = Func(name="a_func", interface=CallInterface(args=("arg",)))
        expected = FileIr(
            context=make_root_context(
                [a_func],
                include_root_symbols=True,
            ),
            file_ir={
                a_func: {
                    "gets": {
                        Name("arg"),
                        Name("arg.b[]", "arg"),
                    },
                    "sets": set(),
                    "dels": {Name("arg.b[].attr", "arg")},
                    "calls": set(),
                },
            },
        )

        assert results == expected

    def test_format_simple(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
        constant: str,
    ):
        ast_ = parse(
            """
            def a_func(arg):
                return format(arg, "b")
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        format_builtin = Builtin(name="format")
        a_func = Func(name="a_func", interface=CallInterface(args=("arg",)))
        format_call = Call(
            name="format()",
            args=CallArguments(args=("arg", constant)),
            target=format_builtin,
        )

        expected = FileIr(
            context=make_root_context(
                [a_func],
                include_root_symbols=True,
            ),
            file_ir={
                a_func: {
                    "gets": {Name("arg")},
                    "sets": set(),
                    "dels": set(),
                    "calls": {format_call},
                },
            },
        )

        assert results == expected

    def test_format_complex(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
        constant: str,
    ):
        ast_ = parse(
            """
            def a_func(arg):
                return format(getattr(arg, "attr"), "b")
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        format_builtin = Builtin(name="format")
        a_func = Func(name="a_func", interface=CallInterface(args=("arg",)))
        format_call = Call(
            name="format()",
            args=CallArguments(args=("arg.attr", constant)),
            target=format_builtin,
        )

        expected = FileIr(
            context=make_root_context(
                [a_func],
                include_root_symbols=True,
            ),
            file_ir={
                a_func: {
                    "gets": {Name("arg"), Name("arg.attr", "arg")},
                    "sets": set(),
                    "dels": set(),
                    "calls": {format_call},
                },
            },
        )

        assert results == expected

    def test_lambda(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
    ):
        ast_ = parse(
            """
            global_lamb = lambda x: x.attr

            def func_one(arg):
                return global_lamb(arg)

            def func_two(arg):
                return map(lambda x: x*x, [1, 2, 3])
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        func_one = Func(name="func_one", interface=CallInterface(args=("arg",)))
        func_two = Func(name="func_two", interface=CallInterface(args=("arg",)))
        global_lamb = Func(name="global_lamb", interface=CallInterface(args=("x",)))
        global_lamb_call = Call(
            name="global_lamb()",
            args=CallArguments(args=("arg",)),
            target=global_lamb,
        )
        map_symbol = Builtin("map")
        map_symbol_call = Call(
            name="map()",
            args=CallArguments(args=("@Lambda", "@List")),
            target=map_symbol,
        )

        expected = FileIr(
            context=make_root_context(
                [
                    global_lamb,
                    func_one,
                    func_two,
                ],
                include_root_symbols=True,
            ),
            file_ir={
                global_lamb: {
                    "gets": {Name("x.attr", "x")},
                    "sets": set(),
                    "dels": set(),
                    "calls": set(),
                },
                func_one: {
                    "gets": {Name("arg")},
                    "sets": set(),
                    "dels": set(),
                    "calls": {global_lamb_call},
                },
                func_two: {
                    "gets": {Name("x")},
                    "sets": set(),
                    "dels": set(),
                    "calls": {map_symbol_call},
                },
            },
        )

        assert results == expected

    def test_class_init(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
    ):
        ast_ = parse(
            """
            class ClassName:
                def __init__(self, arg):
                    self.attr = arg

            def a_func(blarg):
                thing = ClassName(blarg)
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        cls = Class(name="ClassName", interface=CallInterface(args=("self", "arg")))
        cls_call = Call(
            "ClassName()",
            args=CallArguments(args=("thing", "blarg")),
            target=cls,
        )
        a_func = Func(name="a_func", interface=CallInterface(args=("blarg",)))

        expected = FileIr(
            context=make_root_context(
                [
                    cls,
                    a_func,
                ],
                include_root_symbols=True,
            ),
            file_ir={
                cls: {
                    "sets": {Name("self.attr", "self")},
                    "gets": {Name("arg")},
                    "dels": set(),
                    "calls": set(),
                },
                a_func: {
                    "sets": {Name("thing")},
                    "gets": {Name("blarg")},
                    "dels": set(),
                    "calls": {cls_call},
                },
            },
        )

        assert results == expected

    def test_return_implicit_none(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
    ):
        ast_ = parse(
            """
            def a_func(blarg):
                return
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        a_func = Func(name="a_func", interface=CallInterface(args=("blarg",)))
        expected = FileIr(
            context=make_root_context(
                [a_func],
                include_root_symbols=True,
            ),
            file_ir={
                a_func: {
                    "sets": set(),
                    "gets": set(),
                    "dels": set(),
                    "calls": set(),
                },
            },
        )

        assert results == expected

    def test_return_literal(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
    ):
        ast_ = parse(
            """
            def a_func(blarg):
                return 4
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        a_func = Func(name="a_func", interface=CallInterface(args=("blarg",)))
        expected = FileIr(
            context=make_root_context(
                [a_func],
                include_root_symbols=True,
            ),
            file_ir={
                a_func: {
                    "sets": set(),
                    "gets": set(),
                    "dels": set(),
                    "calls": set(),
                },
            },
        )

        assert results == expected

    def test_return_local_variable(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
    ):
        ast_ = parse(
            """
            def a_func(blarg):
                return blarg.attr
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        a_func = Func(name="a_func", interface=CallInterface(args=("blarg",)))
        expected = FileIr(
            context=make_root_context(
                [a_func],
                include_root_symbols=True,
            ),
            file_ir={
                a_func: {
                    "sets": set(),
                    "gets": {Name("blarg.attr", "blarg")},
                    "dels": set(),
                    "calls": set(),
                },
            },
        )

        assert results == expected

    def test_return_tuple_without_class_call(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
    ):
        ast_ = parse(
            """
            def a_func(blarg):
                return blarg.attr, 1, a_call(blarg.another_attr)
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        a_func = Func(name="a_func", interface=CallInterface(args=("blarg",)))
        a_call = Call(
            name="a_call",
            args=CallArguments(args=("blarg.another_attr",)),
            target=None,
        )

        expected = FileIr(
            context=make_root_context(
                [a_func],
                include_root_symbols=True,
            ),
            file_ir={
                a_func: {
                    "sets": set(),
                    "gets": {
                        Name("blarg.attr", "blarg"),
                        Name("blarg.another_attr", "blarg"),
                    },
                    "dels": set(),
                    "calls": {a_call},
                },
            },
        )

        assert results == expected

    def test_return_class_call(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
        constant: str,
    ):
        ast_ = parse(
            """
            class MyEnum(Enum):
                first = "one"
                second = "two"

            def a_func(blarg):
                return MyEnum("one")
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        a_func = Func(name="a_func", interface=CallInterface(args=("blarg",)))
        my_enum_symbol = Class(
            name="MyEnum",
            interface=CallInterface(args=("self", "_id")),
        )
        my_enum_first = Name("MyEnum.first", "MyEnum")
        my_enum_second = Name("MyEnum.second", "MyEnum")
        my_enum_call = Call(
            name="MyEnum",
            args=CallArguments(args=("@ReturnValue", constant)),
            target=my_enum_symbol,
        )

        expected = FileIr(
            context=make_root_context(
                [
                    a_func,
                    my_enum_symbol,
                    my_enum_first,
                    my_enum_second,
                ],
                include_root_symbols=True,
            ),
            file_ir={
                a_func: {
                    "sets": set(),
                    "gets": set(),
                    "dels": set(),
                    "calls": {my_enum_call},
                },
                my_enum_symbol: {
                    "sets": set(),
                    "gets": {my_enum_first, my_enum_second},
                    "calls": set(),
                    "dels": set(),
                },
            },
        )

        assert (
            results.context.symbol_table._symbols
            == expected.context.symbol_table._symbols
        )
        assert results == expected

    def test_return_tuple_with_class(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
        constant: str,
    ):
        ast_ = parse(
            """
            class MyEnum(Enum):
                first = "one"
                second = "two"

            def a_func(blarg):
                return 1, MyEnum("one"), blarg.attr
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        a_func = Func(name="a_func", interface=CallInterface(args=("blarg",)))
        my_enum_symbol = Class(
            name="MyEnum",
            interface=CallInterface(args=("self", "_id")),
        )
        my_enum_first = Name("MyEnum.first", "MyEnum")
        my_enum_second = Name("MyEnum.second", "MyEnum")
        my_enum_call = Call(
            name="MyEnum",
            args=CallArguments(args=("@ReturnValue", constant)),
            target=my_enum_symbol,
        )

        expected = FileIr(
            context=make_root_context(
                [
                    a_func,
                    my_enum_symbol,
                    my_enum_first,
                    my_enum_second,
                ],
                include_root_symbols=True,
            ),
            file_ir={
                a_func: {
                    "sets": set(),
                    "gets": {Name("blarg.attr", "blarg")},
                    "dels": set(),
                    "calls": {my_enum_call},
                },
                my_enum_symbol: {
                    "sets": set(),
                    "gets": {my_enum_first, my_enum_second},
                    "calls": set(),
                    "dels": set(),
                },
            },
        )

        assert results == expected

    def test_walrus(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
    ):
        ast_ = parse(
            """
            def fn(arg):
                if (a := arg.attr):
                    return a
                else:
                    return (b := arg.another_attr)
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        fn = Func(name="fn", interface=CallInterface(args=("arg",)))
        expected = FileIr(
            context=make_root_context([fn], include_root_symbols=True),
            file_ir={
                fn: {
                    "sets": {
                        Name("a"),
                        Name("b"),
                    },
                    "gets": {
                        Name("a"),
                        Name("arg.attr", "arg"),
                        Name("arg.another_attr", "arg"),
                    },
                    "dels": set(),
                    "calls": set(),
                },
            },
        )

        assert results == expected

    def test_walrus_basic(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
    ):
        ast_ = parse(
            """
            def fn(arg):
                thing = (a, b := arg.attr)
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        fn = Func(name="fn", interface=CallInterface(args=("arg",)))
        expected = FileIr(
            context=make_root_context([fn], include_root_symbols=True),
            file_ir={
                fn: {
                    "sets": {
                        Name("b"),
                        Name("thing"),
                    },
                    "gets": {
                        Name("a"),
                        Name("arg.attr", "arg"),
                    },
                    "dels": set(),
                    "calls": set(),
                },
            },
        )

        assert results == expected

    def test_walrus_multiple_assignment(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
    ):
        ast_ = parse(
            """
            def fn(arg):
                x, y = (a, b := c)
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        fn = Func(name="fn", interface=CallInterface(args=("arg",)))
        expected = FileIr(
            context=make_root_context([fn], include_root_symbols=True),
            file_ir={
                fn: {
                    "sets": {
                        Name("x"),
                        Name("y"),
                        Name("b"),
                    },
                    "gets": {
                        Name("a"),
                        Name("c"),
                    },
                    "dels": set(),
                    "calls": set(),
                },
            },
        )

        assert results == expected

    def test_walrus_lambda(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
        capfd: pytest.CaptureFixture[str],
    ):
        ast_ = parse(
            """
            def fn(arg):
                other = (name := lambda *a, **k: a.attr)
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        _, stderr = capfd.readouterr()
        assert "unable to unbind lambdas defined in functions" in stderr

        fn = Func(name="fn", interface=CallInterface(args=("arg",)))
        expected = FileIr(
            context=make_root_context([fn], include_root_symbols=True),
            file_ir={
                fn: {
                    "sets": {
                        Name("other"),
                        Name("name"),
                    },
                    "gets": {
                        Name("a.attr", "a"),
                    },
                    "dels": set(),
                    "calls": set(),
                },
            },
        )

        assert results == expected

    def test_walrus_multiple_assign_with_lambda(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
    ):
        ast_ = parse(
            """
            def fn(arg):
                other = (alpha, beta := lambda *a, **k: a.attr)
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        fn = Func(name="fn", interface=CallInterface(args=("arg",)))
        expected = FileIr(
            context=make_root_context([fn], include_root_symbols=True),
            file_ir={
                fn: {
                    "sets": {
                        Name("other"),
                        Name("beta"),
                    },
                    "gets": {
                        Name("alpha"),
                        Name("a.attr", "a"),
                    },
                    "dels": set(),
                    "calls": set(),
                },
            },
        )

        assert results == expected

    def test_walrus_multiple_assign_with_lambda_as_list(
        self,
        parse: ParseFn,
        make_root_context: MakeRootContextFn,
    ):
        ast_ = parse(
            """
            def fn(arg):
                other = [alpha, beta := lambda *a, **k: a.attr]
            """
        )
        results = FileAnalyser(ast_, compile_root_context(ast_)).analyse()

        fn = Func(name="fn", interface=CallInterface(args=("arg",)))
        expected = FileIr(
            context=make_root_context([fn], include_root_symbols=True),
            file_ir={
                fn: {
                    "sets": {
                        Name("other"),
                        Name("beta"),
                    },
                    "gets": {
                        Name("alpha"),
                        Name("a.attr", "a"),
                    },
                    "dels": set(),
                    "calls": set(),
                },
            },
        )

        assert results == expected



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest import mock

import attrs
from attrs import field

from rattr.models.context._root_context import (
    MODULE_LEVEL_DUNDER_ATTRS,
    PYTHON_BUILTINS,
)
from rattr.models.symbol import Import

if TYPE_CHECKING:
    import ast
    from collections.abc import Iterable, Iterator, Mapping
    from pathlib import Path
    from typing import Any, Protocol, TypeVar

    from rattr.ast.types import Identifier
    from rattr.config import Arguments, State
    from rattr.models.context import Context, SymbolTable
    from rattr.models.ir import FileIr, FunctionIr
    from rattr.models.symbol import Symbol, UserDefinedCallableSymbol
    from rattr.module_locator.models import ModuleSpec
    from rattr.module_locator.util import ModuleName

    class FileIrFromDictFn(Protocol):
        def __call__(
            self,
            ir: Mapping[UserDefinedCallableSymbol, FunctionIr],
        ) -> FileIr:
            ...

    class MakeSymbolTableFn(Protocol):
        def __call__(
            self,
            symbols: Mapping[Identifier, Symbol] | Iterable[Symbol],
            *,
            include_root_symbols: bool = False,
        ) -> SymbolTable:
            ...

    class MakeRootContextFn(Protocol):
        def __call__(
            self,
            symbols: Mapping[Identifier, Symbol] | Iterable[Symbol],
            *,
            include_root_symbols: bool = False,
        ) -> Context:
            ...

    class ArgumentsFn(Protocol):
        def __call__(
            self,
            **kwargs: Mapping[str, Any],
        ) -> Iterator[None]:
            ...

    class StateFn(Protocol):
        def __call__(
            self,
            **kwargs: Mapping[str, Any],
        ) -> Iterator[None]:
            ...

    class SetTestingConfigFn(Protocol):
        def __call__(self, arguments: Arguments, state: State = None) -> None:
            ...

    class ParseFn(Protocol):
        def __call__(self, source: str) -> ast.Module:
            ...

    class ParseWithContextFn(Protocol):
        def __call__(self, source: str) -> tuple[ast.Module, Context]:
            ...

    StrOrPath = TypeVar("StrOrPath", str, Path)

    class OsDependentPathFn(Protocol):
        def __call__(self, target: StrOrPath) -> StrOrPath:
            ...


@attrs.frozen
class Import_(Import):
    """
    This is a specialisation of `Import` where the module name and spec can be set
    explicitly, which is useful for many tests.
    """

    _module_name_and_spec: tuple[ModuleName, ModuleSpec] = field(
        alias="module_name_and_spec",
        default=("blah", "blah"),
    )


def compare_identifier_symbol_mappings(
    lhs: Mapping[Identifier, FunctionIr],
    rhs: Mapping[Identifier, FunctionIr],
    /,
) -> bool:
    errors: list[str] = []

    keys = lhs.keys()
    other_keys = rhs.keys()

    if extra := (keys - other_keys):
        errors.append(f"extra keys in lhs: {extra}")
    if extra := (other_keys - keys):
        errors.append(f"extra keys in rhs: {extra}")

    for key in keys & other_keys:
        if key in MODULE_LEVEL_DUNDER_ATTRS or key in PYTHON_BUILTINS:
            lhs_symbol = attrs.evolve(lhs[key], token=mock.ANY, location=mock.ANY)
            rhs_symbol = attrs.evolve(rhs[key], token=mock.ANY, location=mock.ANY)
        else:
            lhs_symbol = lhs[key]
            rhs_symbol = rhs[key]

        if lhs_symbol == rhs_symbol:
            continue

        errors.append(f"{lhs_symbol} != {rhs_symbol}")

    if errors:
        print("\n".join(errors))
        return False

    return True


def compare_symbol_table_symbols(
    lhs: SymbolTable,
    rhs: SymbolTable,
    /,
) -> bool:
    return compare_identifier_symbol_mappings(lhs, rhs)


def match_output(
    actual_stdout_or_stderr: str | Iterable[str],
    expected_stdout_or_stderr: str | Iterable[str],
) -> bool:
    if isinstance(actual_stdout_or_stderr, str):
        actual_lines = actual_stdout_or_stderr.splitlines()
    else:
        actual_lines = actual_stdout_or_stderr

    if isinstance(expected_stdout_or_stderr, str):
        expected_lines = expected_stdout_or_stderr.splitlines()
    else:
        expected_lines = expected_stdout_or_stderr

    has_error: bool = False

    if len(actual_lines) != len(expected_lines):
        has_error = True

    for actual, expected in zip(actual_lines, expected_lines):
        # TODO
        # We check endswith to avoid the ANSI chars, but this could be better
        if not actual.endswith(expected):
            has_error = True

    if has_error:
        print("actual output:")
        print("\n".join(actual_lines))
        print("----------")
        print("expected output:")
        print("\n".join(expected_lines))
        return False

    return True



"""CustomFunctionAnalyser for Python's builtin functions."""
from __future__ import annotations

import ast
from typing import TYPE_CHECKING, NamedTuple

from rattr.analyser.base import CustomFunctionAnalyser
from rattr.analyser.util import get_dynamic_name
from rattr.ast.util import fullname_of
from rattr.models.context import Context
from rattr.models.symbol import Name
from rattr.results import unbind_ir_with_call_swaps

if TYPE_CHECKING:
    from collections.abc import Iterator

    from rattr.ast.types import Identifier
    from rattr.models.ir import FunctionIr


class AccessedAttributes(NamedTuple):
    full: Name
    lhs_names: list[Name]


def iter_lhs_names(target: Name) -> Iterator[Identifier]:
    parts = target.name.split(".")

    if parts == [target]:
        yield target
        return

    for end_offset in range(1, len(parts)):
        yield ".".join(parts[:-end_offset])


def accessed_attributes(name: str, node: ast.Call) -> AccessedAttributes:
    return AccessedAttributes(
        full=(full := get_dynamic_name(name, node, "{first}.{second}")),
        lhs_names=[Name(name) for name in iter_lhs_names(full)],
    )


class GetattrAnalyser(CustomFunctionAnalyser):
    @property
    def name(self) -> str:
        return "getattr"

    @property
    def qualified_name(self) -> str:
        return "getattr"

    def on_def(
        self,
        name: str,
        node: ast.FunctionDef | ast.AsyncFunctionDef,
        ctx: Context,
    ) -> FunctionIr:
        return super().on_def(name, node, ctx)

    def on_call(self, name: str, node: ast.Call, ctx: Context) -> FunctionIr:
        full, lhs_names = accessed_attributes(name, node)
        return {
            "gets": {full, *lhs_names},
            "sets": set(),
            "dels": set(),
            "calls": set(),
        }


class SetattrAnalyser(CustomFunctionAnalyser):
    @property
    def name(self) -> str:
        return "setattr"

    @property
    def qualified_name(self) -> str:
        return "setattr"

    def on_def(
        self,
        name: str,
        node: ast.FunctionDef | ast.AsyncFunctionDef,
        ctx: Context,
    ) -> FunctionIr:
        return super().on_def(name, node, ctx)

    def on_call(self, name: str, node: ast.Call, ctx: Context) -> FunctionIr:
        full, lhs_names = accessed_attributes(name, node)
        return {
            "gets": {*lhs_names},
            "sets": {full},
            "dels": set(),
            "calls": set(),
        }


class HasattrAnalyser(CustomFunctionAnalyser):
    @property
    def name(self) -> str:
        return "hasattr"

    @property
    def qualified_name(self) -> str:
        return "hasattr"

    def on_def(
        self,
        name: str,
        node: ast.FunctionDef | ast.AsyncFunctionDef,
        ctx: Context,
    ) -> FunctionIr:
        return super().on_def(name, node, ctx)

    def on_call(self, name: str, node: ast.Call, ctx: Context) -> FunctionIr:
        full, lhs_names = accessed_attributes(name, node)
        return {
            "gets": {full, *lhs_names},
            "sets": set(),
            "dels": set(),
            "calls": set(),
        }


class DelattrAnalyser(CustomFunctionAnalyser):
    @property
    def name(self) -> str:
        return "delattr"

    @property
    def qualified_name(self) -> str:
        return "delattr"

    def on_def(
        self,
        name: str,
        node: ast.FunctionDef | ast.AsyncFunctionDef,
        ctx: Context,
    ) -> FunctionIr:
        return super().on_def(name, node, ctx)

    def on_call(self, name: str, node: ast.Call, ctx: Context) -> FunctionIr:
        full, lhs_names = accessed_attributes(name, node)
        return {
            "gets": {*lhs_names},
            "sets": set(),
            "dels": {full},
            "calls": set(),
        }


class SortedAnalyser(CustomFunctionAnalyser):
    @property
    def name(self) -> str:
        return "sorted"

    @property
    def qualified_name(self) -> str:
        return "sorted"

    def on_def(
        self,
        name: str,
        node: ast.FunctionDef | ast.AsyncFunctionDef,
        ctx: Context,
    ) -> FunctionIr:
        return super().on_def(name, node, ctx)

    def on_call(self, name: str, node: ast.Call, ctx: Context) -> FunctionIr:
        # HACK Avoid circular import
        from rattr.analyser.function import FunctionAnalyser

        if len(node.args) == 0:
            return super().on_call(name, node, ctx)

        fn_analyser = FunctionAnalyser(ast.FunctionDef(), ctx)

        # The first arg may be a get/call within the context
        fn_analyser.visit(node.args[0])

        # If there is a key, it should be visited
        key = None

        for kw in node.keywords:
            if kw.arg == "key":
                key = kw

        if key is None:
            return fn_analyser.func_ir

        if isinstance(key.value, ast.Lambda):
            # NOTE The lambda in `sorted` expects exactly one argument
            if len(key.value.args.args) != 1:
                raise SyntaxError

            # HACK
            #   Visit the lambda in a different FunctionAnalyser s.t. the
            #   key-argument-name can be unbound without interfering with the
            #   results from visiting the sorting target

            iterator = key.value.args.args[0].arg
            iterable = fullname_of(node.args[0], safe=True)

            lambda_ctx = Context(ctx)
            lambda_ctx.add(Name(iterator, token=node), is_argument=True)

            lambda_analyser = FunctionAnalyser(ast.FunctionDef(), lambda_ctx)
            lambda_analyser.visit(key.value.body)

            # Substitute
            lambda_analyser.func_ir = unbind_ir_with_call_swaps(
                lambda_analyser.func_ir,
                {iterator: iterable},
            )

            fn_analyser.func_ir["gets"] |= lambda_analyser.func_ir["gets"]
            fn_analyser.func_ir["sets"] |= lambda_analyser.func_ir["sets"]
            fn_analyser.func_ir["dels"] |= lambda_analyser.func_ir["dels"]
            fn_analyser.func_ir["calls"] |= lambda_analyser.func_ir["calls"]

        else:
            fn_analyser.visit(key.value)

        return fn_analyser.func_ir



from __future__ import annotations

import argparse
import re
from dataclasses import dataclass, field
from enum import Enum, IntFlag, auto
from functools import cached_property, lru_cache
from pathlib import Path
from typing import TYPE_CHECKING

from rattr.config._util import (
    find_project_root,
    find_pyproject_toml,
    validate_arguments,
)

if TYPE_CHECKING:
    from typing import Final, Literal


@lru_cache(maxsize=None)
def _cached_re_compile(pattern: str) -> re.Pattern[str]:
    return re.compile(pattern)


class FollowImports(IntFlag):
    local = auto()
    pip = auto()
    stdlib = auto()


class ShowWarnings(IntFlag):
    target = auto()
    target_low_priority = auto()
    inherited_high_priority = auto()
    inherited_low_priority = auto()


class FormatPath(IntFlag):
    collapse_home = auto()
    truncate_deep_paths = auto()


class Output(Enum):
    stats = "stats"
    ir = "ir"
    results = "results"
    cacheable = "cacheable"
    silent = "silent"

    def __str__(self) -> str:
        return self.name


class Arguments(argparse.Namespace):
    pyproject_toml_override: Path | None
    """From `[-c PATH | --config PATH]`."""

    _follow_imports_level: Literal[0, 1, 2, 3]

    _excluded_imports: list[str] | None
    _excluded_names: list[str] | None

    _warning_level: Literal["none", "local", "default", "all"]

    collapse_home: bool
    truncate_deep_paths: bool

    is_strict: bool
    threshold: int

    stdout: Output

    target: Path

    @property
    def follow_imports(self) -> FollowImports:
        if self._follow_imports_level == 0:
            return FollowImports(0)
        if self._follow_imports_level == 1:
            return FollowImports.local
        if self._follow_imports_level == 2:
            return FollowImports.local | FollowImports.pip
        if self._follow_imports_level == 3:
            return FollowImports.local | FollowImports.pip | FollowImports.stdlib
        raise NotImplementedError

    @property
    def follow_local_imports(self) -> bool:
        return FollowImports.local in self.follow_imports

    @property
    def follow_pip_imports(self) -> bool:
        return FollowImports.pip in self.follow_imports

    @property
    def follow_stdlib_imports(self) -> bool:
        return FollowImports.stdlib in self.follow_imports

    @property
    def excluded_imports(self) -> set[str]:
        if self._excluded_imports is None:
            return set()
        return set(self._excluded_imports)

    @property
    def excluded_names(self) -> set[str]:
        if self._excluded_names is None:
            return set()
        return set(self._excluded_names)

    @property
    def re_excluded_names(self) -> tuple[re.Pattern[str], ...]:
        return tuple(_cached_re_compile(p) for p in self.excluded_names)

    @property
    def show_warnings(self) -> ShowWarnings:
        if self._warning_level == "none":
            return ShowWarnings(0)
        if self._warning_level == "local":
            return ShowWarnings.target
        if self._warning_level == "default":
            return ShowWarnings.target | ShowWarnings.inherited_high_priority
        if self._warning_level == "all":
            return (
                ShowWarnings.target
                | ShowWarnings.target_low_priority
                | ShowWarnings.inherited_high_priority
                | ShowWarnings.inherited_low_priority
            )
        raise NotImplementedError

    @property
    def format_path(self) -> FormatPath:
        flag = FormatPath(0)

        if self.collapse_home:
            flag |= FormatPath.collapse_home
        if self.truncate_deep_paths:
            flag |= FormatPath.truncate_deep_paths

        return flag


@dataclass
class State:
    badness_from_target_file: int = 0
    badness_from_imports: int = 0
    badness_from_simplification: int = 0

    current_file: "Path | None" = None

    @property
    def is_in_any_file(self) -> bool:
        return self.current_file is not None

    @property
    def badness(self) -> int:
        return self.badness_from_target_file + self.badness_from_simplification

    @property
    def full_badness(self) -> int:
        return (
            self.badness_from_target_file
            + self.badness_from_imports
            + self.badness_from_simplification
        )


class ConfigMetaclass(type):
    """Metaclass allowing `Config` to act as a singleton."""

    _instance: Config | None = None

    def __call__(cls, *args, **kwargs):
        if cls._instance is None:
            _instance: Config = super().__call__(*args, **kwargs)
            cls._instance = _instance
            cls._instance.arguments = validate_arguments(cls._instance.arguments)
        return cls._instance


@dataclass
class Config(metaclass=ConfigMetaclass):
    """The global config singleton."""

    arguments: Arguments
    state: State

    LITERAL_VALUE_PREFIX: Final = "@"
    """The prefix given to local constants, literals, etc to produce a name."""

    MODULE_BLACKLIST_PATTERNS: Final = frozenset(
        (
            r"rattr",
            r"rattr\..*",
            r"packages?\.rattr",
            r"packages?\.rattr\..*",
        )
    )
    """The set of perennial module blacklist patterns, i.e. rattr itself."""

    PLUGINS_BLACKLIST_PATTERNS: set[str] = field(default_factory=set)
    """The set of blacklist patterns set by custom plugins."""

    @cached_property
    def project_root(self) -> Path:
        return find_project_root()

    @cached_property
    def pyproject_toml(self) -> Path | None:
        return find_pyproject_toml()

    @cached_property
    def root_cache_dir(self) -> Path:
        return self.project_root / ".rattr" / "cache"

    @property
    def is_in_target_file(self) -> bool:
        return self.arguments.target == self.state.current_file

    @property
    def do_not_follow_imports(self) -> bool:
        return self.arguments.follow_imports == FollowImports(0)

    @property
    def do_not_show_warnings(self) -> bool:
        return self.arguments.show_warnings == ShowWarnings(0)

    @property
    def use_full_path(self) -> bool:
        return self.arguments.format_path == FormatPath(0)

    def increment_badness(self, badness: int) -> None:
        """Increment the appropriate badness level based on state."""
        if badness < 0:
            raise ValueError("'badness' must be positive integer")

        if not self.state.is_in_any_file:
            self.state.badness_from_simplification += badness
        elif self.is_in_target_file:
            self.state.badness_from_target_file += badness
        else:
            self.state.badness_from_imports += badness

    @property
    def is_within_badness_threshold(self) -> int:
        if self.arguments.is_strict:
            return self.state.badness <= 0

        # A threshold of zero is equivalent to infinite
        if self.arguments.threshold == 0:
            return True

        return self.state.badness <= self.arguments.threshold

    def get_formatted_path(self, path: Path | str | None) -> str | None:
        if path is None:
            return None

        if isinstance(path, str):
            path = Path(path)

        home = str(Path.home().resolve())

        if path.is_relative_to(self.project_root):
            path = path.relative_to(self.project_root)

        if self.arguments.collapse_home and path.is_relative_to(home):
            relative_path = str(path).replace(home, "")

            if relative_path.startswith(("/", "\\")):
                relative_path = relative_path[1:]

            path = Path("~") / relative_path

        if self.arguments.truncate_deep_paths and len(path.parts) > 5:
            base = Path(path.parts[0]) / "..."

            for part in path.parts[-3:]:
                base /= part

            path = base

        return path.as_posix()

    @property
    def formatted_current_file_path(self) -> str | None:
        return self.get_formatted_path(self.state.current_file)

    @property
    def formatted_target_path(self) -> str | None:
        return self.get_formatted_path(self.arguments.target)

    @property
    def blacklist_patterns(self) -> set[str]:
        return (
            self.arguments.excluded_imports
            | self.MODULE_BLACKLIST_PATTERNS
            | self.PLUGINS_BLACKLIST_PATTERNS
        )

    @property
    def re_blacklist_patterns(self) -> tuple[re.Pattern[str], ...]:
        return tuple(_cached_re_compile(p) for p in self.blacklist_patterns)



from __future__ import annotations

from typing import TYPE_CHECKING

import attrs
from attrs import field

from rattr.analyser.util import get_file_hash
from rattr.models.results import FileResults
from rattr.models.symbol import Import
from rattr.models.symbol._util import PYTHON_BUILTINS_LOCATION
from rattr.module_locator.util import is_in_import_blacklist

if TYPE_CHECKING:
    from rattr.analyser.types import ImportIrs
    from rattr.models.ir import FileIr


def make_cacheable_results(
    results: FileResults,
    target_ir: FileIr,
    import_irs: ImportIrs,
) -> CacheableResults:
    return CacheableResults(
        filepath=target_ir.context.file,
        filehash=get_file_hash(target_ir.context.file),
        imports=make_cacheable_import_info(target_ir, import_irs),
        results=results,
    )


def make_cacheable_import_info(
    target_ir: FileIr,
    import_irs: ImportIrs,
) -> list[CacheableImportInfo]:
    contexts = (
        target_ir.context,
        *(import_.context for import_ in import_irs.values()),
    )

    return sorted(
        {
            CacheableImportInfo.from_file(symbol.module_spec.origin)
            for context in contexts
            for symbol in context.symbol_table.symbols
            if isinstance(symbol, Import)
            if symbol.module_name is not None
            if not is_in_import_blacklist(symbol.module_name)
            if symbol.module_spec is not None
            if symbol.module_spec.origin is not None
            if symbol.module_spec.origin != PYTHON_BUILTINS_LOCATION
        },
        key=lambda info: info.filepath,
    )


@attrs.frozen
class CacheableImportInfo:
    filepath: str = field(converter=str)
    filehash: str

    @classmethod
    def from_file(cls, filepath: str) -> CacheableImportInfo:
        return CacheableImportInfo(filepath=filepath, filehash=get_file_hash(filepath))


@attrs.frozen
class CacheableResults:
    filepath: str = field(converter=str)
    filehash: str
    imports: list[CacheableImportInfo]
    results: FileResults



from __future__ import annotations

from rattr.plugins.analysers.builtins import (
    DelattrAnalyser,
    GetattrAnalyser,
    HasattrAnalyser,
    SetattrAnalyser,
    SortedAnalyser,
)
from rattr.plugins.analysers.collections import DefaultDictAnalyser

DEFAULT_FUNCTION_ANALYSERS = (
    GetattrAnalyser(),
    SetattrAnalyser(),
    HasattrAnalyser(),
    DelattrAnalyser(),
    SortedAnalyser(),
    DefaultDictAnalyser(),
)



from __future__ import annotations

from rattr.models.symbol._util import (  # noqa: F401
    PYTHON_BUILTINS_LOCATION,
    arg_name,
    kwarg_name,
)


def with_call_brackets(call: str) -> str:
    """Return the given string with call brackets added if absent.

    >>> without_call_brackets("keep_my_call_brackets()")
    "keep_my_call_brackets()"
    >>> without_call_brackets("i_dont_have_call_brackets")
    "i_dont_have_call_brackets()"
    """
    if not call.endswith("()"):
        return f"{call}()"

    return call


def without_call_brackets(call: str) -> str:
    """Return the given string with call brackets removed if present.

    >>> without_call_brackets("remove_my_call_brackets()")
    "remove_my_call_brackets"
    >>> without_call_brackets("i_dont_have_call_brackets")
    "i_dont_have_call_brackets"
    """
    while call.endswith("()"):
        call = call.removesuffix("()")
    return call


def get_basename_from_name(name: str) -> str:
    """Return the basename for the given name.

    >>> get_basename_from_name("my_var")
    "my_var"

    >>> get_basename_from_name("my_var.attr")
    "my_var"

    >>> get_basename_from_name("*my_var.attr.method()")
    "my_var"
    """
    return without_call_brackets(name).replace("*", "").split(".", maxsplit=1)[0]



from __future__ import annotations

import ast
import re
from typing import TYPE_CHECKING, TypedDict

from rattr import error
from rattr.config import Config
from rattr.models.symbol import Name

if TYPE_CHECKING:
    from typing import Final, Literal

    from rattr.ast.types import Identifier
    from rattr.models.ir import FunctionIr
    from rattr.models.symbol import Call, Func


VARARG_NAME: Final = f"{Config.LITERAL_VALUE_PREFIX}{ast.Tuple.__name__}"
KWARGS_NAME: Final = f"{Config.LITERAL_VALUE_PREFIX}{ast.Dict.__name__}"


def construct_call_swaps(func: Func, call: Call) -> dict[Identifier, Identifier]:
    """Return the map of local function names to bound names."""
    # TODO
    # The func.interface needs to store the default arguments, the ast makes this tricky
    # as it is just a list of default values (not a dict of argument to default value)
    # but as you can't have a non-defaulted value after a defaulted one we can work this
    # out at when we build the context.
    # TODO
    # Handle vararg and kwargs
    swaps: dict[Identifier, Identifier] = {}

    # Copy the input args / interface so we can pop to keep track of state
    interface = func.interface.as_consumable_call_interface()
    call_args = [a for a in call.args.args]
    call_kwargs = {k: v for k, v in call.args.kwargs.items()}

    while True:
        if not interface.posonlyargs:
            break

        if not call_args:
            error.error(
                f"call to {func.name!r} expected {len(func.interface.posonlyargs)} "
                f"posonlyargs but only received {len(call.args.args)} positional "
                f"arguments",
                culprit=call,
            )
            return {}

        target = interface.posonlyargs.pop(0)
        replacement = call_args.pop(0)

        swaps[target] = replacement

    while True:
        if not interface.args:
            break  # no error as the remaining args may be supplied by kw

        if not call_args:
            break

        target = interface.args.pop(0)
        replacement = call_args.pop(0)

        swaps[target] = replacement

    if interface.vararg is not None:
        swaps[interface.vararg] = VARARG_NAME
        call_args = []

    # We can't be very specific here as we don't know if any of the positional arguments
    # have defaults (yet, we need a better CallInterface, see the todo above).
    if call_args:
        error.error(
            f"call to {func.name!r} received too many positional arguments",
            culprit=call,
        )

    # interface.args may be [] or [...]
    # call_args is []

    unexpected_keyword_arguments: list[Identifier] = []
    arguments_given_by_position_and_name: list[Identifier] = []

    while True:
        if not call_kwargs:
            break

        (target, replacement) = call_kwargs.popitem()

        if target in swaps:
            arguments_given_by_position_and_name.append(target)

        if target in interface.args:
            interface.args.remove(target)
            swaps[target] = replacement
        elif target in interface.kwonlyargs:
            interface.kwonlyargs.remove(target)
            swaps[target] = replacement
        elif interface.kwarg is not None:
            swaps[interface.kwarg] = KWARGS_NAME
        elif target not in func.interface.all:
            unexpected_keyword_arguments.append(target)

    if unexpected_keyword_arguments:
        error.error(
            f"call to {func.name!r} received unexpected keyword arguments: "
            f"{unexpected_keyword_arguments}",
            culprit=call,
        )

    if arguments_given_by_position_and_name:
        error.error(
            f"call to {func.name!r} received the arguments "
            f"{arguments_given_by_position_and_name} by position and name",
            culprit=call,
        )

    # There could be interface.args remaining but for now we don't know if they might
    # be defaulted so we can't do anything here
    # There could also be interface.kwonlyargs but, likewise, they could have defaults

    return swaps


def unbind_ir_with_call_swaps(
    ir: FunctionIr,
    swaps: dict[Identifier, Identifier],
) -> FunctionIr:
    return {
        "gets": {unbind_name(n, swaps.get(n.basename, n.basename)) for n in ir["gets"]},
        "sets": {unbind_name(n, swaps.get(n.basename, n.basename)) for n in ir["sets"]},
        "dels": {unbind_name(n, swaps.get(n.basename, n.basename)) for n in ir["dels"]},
        "calls": ir["calls"],
    }


def unbind_name(symbol: Name, new_basename: Identifier) -> Name:
    """Return a new symbol bound to the new base name."""
    if symbol.basename == new_basename:
        return symbol

    if symbol.name.startswith("*"):
        old, new = f"*{symbol.basename}", f"*{new_basename}"
    else:
        old, new = symbol.basename, new_basename

    if not symbol.name.startswith(old):
        # If this is ever true then we'd need to do a regex over:
        # r"^(?P<star>\*)?(?P<basename>\w+)(?P<remainder>(\.\w+)*)(?P<subscript>\[\])?$"
        # constructing a new string from the match groups with basename replaced with
        # the new_basename if it matches the old basename or else returning the target.
        # However, this is much faster (0.25s for 1M vs 0.71s for 1M) and the assumption
        # holds.
        raise ValueError("never")

    new_name = symbol.name.replace(old, new, 1)
    return Name(name=new_name, basename=new_basename, location=symbol.location)


re_name = re.compile(
    r"^(?P<star>\*)?"
    r"(?P<basename>\w+)"
    r"(?P<remainder>(\.\w+)*)"
    r"(?P<subscript>\[\])?$"
)


class ReNameMatches(TypedDict):
    star: Literal["*"] | None
    basename: str
    remainder: str
    subscript: Literal["[]"] | None


def fast_re_replace(target: Identifier, old: Identifier, new: Identifier) -> Identifier:
    matches: ReNameMatches = re_name.fullmatch(target)

    if matches is None:
        return target
    if matches["basename"] != old:
        return target

    star = matches["star"] or ""
    subscript = matches["subscript"] or ""

    return f"{star}{new}{matches['remainder']}{subscript}"



from __future__ import annotations

from collections.abc import MutableMapping
from typing import TYPE_CHECKING

import attrs
from attrs import field

from rattr.ast.types import Identifier
from rattr.models.symbol import Symbol

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, KeysView, ValuesView


@attrs.mutable
class SymbolTable(MutableMapping[Identifier, Symbol]):
    _symbols: dict[Identifier, Symbol] = field(init=False, factory=dict)

    @property
    def names(self) -> KeysView[Identifier]:
        return self._symbols.keys()

    @property
    def symbols(self) -> ValuesView[Symbol]:
        return self._symbols.values()

    def add(self, symbol_or_symbols: Symbol | Iterable[Symbol]) -> None:
        """Add the given symbol(s) to the symbol table."""
        if isinstance(symbol_or_symbols, Symbol):
            symbols = [symbol_or_symbols]
        else:
            symbols = symbol_or_symbols

        for symbol in symbols:
            self[symbol.id] = symbol

    def remove(
        self,
        target_or_targets: Identifier | Symbol | Iterable[Identifier | Symbol],
    ) -> None:
        """Remove the given target(s) from the symbol table."""
        if isinstance(target_or_targets, (Identifier, Symbol)):
            targets = [target_or_targets]
        else:
            targets = target_or_targets

        for t in targets:
            del self[t if isinstance(t, Identifier) else t.id]

    def pop(self, target: Identifier | Symbol) -> Symbol | None:
        """Return and remove the given target, returns `None` if absent."""
        id = target.id if isinstance(target, Symbol) else target

        if id not in self:
            return

        symbol = self[id]
        del self[id]
        return symbol

    # ================================================================================ #
    # Mutable mapping abstract methods and mixin-overrides
    # ================================================================================ #

    def __getitem__(self, __key: Identifier) -> Symbol:
        return self._symbols.__getitem__(__key)

    def __setitem__(self, __key: Identifier, __value: Symbol) -> None:
        return self._symbols.__setitem__(__key, __value)

    def __delitem__(self, __key: Identifier) -> None:
        return self._symbols.__delitem__(__key)

    def __iter__(self) -> Iterator[Identifier]:
        return self._symbols.__iter__()

    def __len__(self) -> int:
        return self._symbols.__len__()

    def clear(self) -> None:
        # Faster than MutableMapping's default clear implementation
        return self._symbols.clear()



from __future__ import annotations

from typing import Union

from rattr.models.symbol._symbols import Builtin, Class, Func, Import
from rattr.versioning.typing import TypeAlias

CallableSymbol: TypeAlias = Union[Builtin, Class, Func, Import]
UserDefinedCallableSymbol: TypeAlias = Union[Class, Func]



from __future__ import annotations


class RattrResultsError(Exception):
    def __init__(self, message: str = "invalid rattr results") -> None:
        self.message = message
        super().__init__(message)



from __future__ import annotations

from typing import TYPE_CHECKING

from rattr import error
from rattr.analyser.util import is_excluded_name
from rattr.config import Config
from rattr.models.symbol import Builtin, Call, Class, Func, Import, Name
from rattr.module_locator.util import (
    derive_module_name_from_path,
    is_in_import_blacklist,
    is_in_pip,
    is_in_stdlib,
)
from rattr.results import IrTarget

if TYPE_CHECKING:
    from rattr.ast.types import Identifier
    from rattr.results import IrCall, IrEnvironment


def find_call_target_and_ir(
    call: IrCall,
    *,
    environment: IrEnvironment,
) -> IrTarget | None:
    target = call.symbol.target

    # TODO Should this trigger a warning?
    if target is None:
        return None

    if isinstance(target, Builtin):
        return None

    # Procedural parameter, etc: can't be resolved
    if isinstance(target, Name):
        return None

    if isinstance(target, Func):
        return resolve_function(call, environment=environment)

    if isinstance(target, Class):
        return resolve_class_init(call, environment=environment)

    if isinstance(target, Import):
        return resolve_import(target, environment=environment)

    raise TypeError(f"{call} is not an IrCall")


def resolve_function(
    call: IrCall,
    *,
    environment: IrEnvironment,
) -> IrTarget | None:
    if call.symbol.target is None:
        raise ImportError

    error_ = f"unable to resolve call to {call.symbol.target.name!r}"

    if call.caller is not None:
        error_ = f"{error_} in {call.caller.name!r}"

    if is_excluded_name(call.symbol.target.name):
        error.error(f"{error_}, the target matches an exclusion", culprit=call.symbol)
        return None

    try:
        target = __resolve_target_and_ir(call, environment=environment)
    except ImportError:
        if not is_call_to_method_or_member(call.symbol):
            error_ += ", the target is likely a nested function or @rattr_ignore'd"
            error.error(f"{error_}", culprit=call.symbol)
        else:
            error.info(f"{error_}", culprit=call.symbol)
        return None

    return target


def resolve_class_init(
    call: IrCall,
    *,
    environment: IrEnvironment,
) -> IrTarget | None:
    error_ = f"unable to resolve initialiser for {call.symbol.target.name!r}"

    try:
        target = __resolve_target_and_ir(call, environment=environment)
    except ImportError:
        error.error(error_, culprit=call.symbol)
        return None

    return target


def resolve_import(
    target: Import,
    *,
    environment: IrEnvironment,
) -> IrTarget | None:
    config = Config()
    arguments = config.arguments

    module_ = f"{target.module_name!r}"
    error_ = f"ignoring call to {target.name!r} imported from {{loc}}"

    if target.module_name is None:
        raise ImportError

    if is_in_import_blacklist(target.module_name):
        return None

    if not arguments.follow_local_imports:
        error.info(error_.format(loc="local module"), culprit=target)
        return None

    if not arguments.follow_pip_imports and is_in_pip(target.module_name):
        error.info(error_.format(loc=f"pip installed module {module_}"), culprit=target)
        return None

    if not arguments.follow_stdlib_imports and is_in_stdlib(target.module_name):
        error.info(error_.format(loc=f"stdlib module {module_}"), culprit=target)
        return None

    module_ir = environment.import_irs.get(target.module_name, None)

    if module_ir is None:
        raise ImportError(f"{module_} not found")

    local_name = target.name.replace(f"{target.module_name}.", "").removesuffix("()")
    new_target = module_ir.context.get(local_name)

    imported_as_ = (
        f" (imported as {target.name!r})" if target.name != local_name else ""
    )
    unresolved_ = (
        f"unable to resolve call to {local_name!r} in import {module_}{imported_as_}, "
        f"{{why}}"
    )

    if isinstance(new_target, (Func, Class)):
        ir = module_ir.get(new_target)

        # NOTE
        # If the imported function is ignored then it will have no IR
        if ir is None:
            error.error(unresolved_.format(why="it is likely ignored"), culprit=target)
            return None

        return IrTarget(symbol=new_target, ir=ir)

    if isinstance(new_target, Import):
        return resolve_import(new_target, environment=environment)

    if new_target is None and is_call_to_method_or_member(local_name):
        error.info(unresolved_.format(why="it is a method"), culprit=target)
    else:
        error.error(unresolved_.format(why="it is likely undefined"), culprit=target)

    return None


def __resolve_target_and_ir(
    call: IrCall,
    *,
    environment: IrEnvironment,
) -> IrTarget:
    if isinstance(call.symbol.target, Class):
        symbol = __resolve_real_class_target(
            call.symbol.target,
            environment=environment,
        )
    else:
        symbol = call.symbol.target

    if symbol is None:
        raise ImportError

    if symbol in environment.target_ir:
        return IrTarget(symbol=symbol, ir=environment.target_ir[symbol])

    filename = symbol.location.defined_in
    module = derive_module_name_from_path(filename)

    if module is None:
        raise ModuleNotFoundError(f"unable to find module for {str(filename)!r}")

    module_ir = environment.import_irs.get(module)

    if module_ir is None:
        raise ImportError
    if symbol not in module_ir:
        raise ImportError

    return IrTarget(symbol=symbol, ir=module_ir[symbol])


def __resolve_real_class_target(
    target: Class,
    *,
    environment: IrEnvironment,
) -> Class:
    for symbol in environment.target_ir:
        if isinstance(symbol, Class) and target.name == symbol.name:
            return symbol

    for _, import_ir in environment.import_irs.items():
        for symbol in import_ir:
            if isinstance(symbol, Class) and target.name == symbol.name:
                return symbol

    return target


def is_call_to_method_or_member(target: Call | Identifier) -> bool:
    if isinstance(target, Call):
        target = target.id
    return "." in target



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from rattr.models.ir import FileIr
from rattr.models.symbol import (
    Call,
    CallArguments,
    CallInterface,
    Class,
    Func,
    Import,
    Name,
)
from rattr.results._types import IrEnvironment

if TYPE_CHECKING:
    from collections.abc import Iterator

    from tests.shared import MakeRootContextFn, StateFn


@pytest.fixture(autouse=True)
def __set_current_file(state: StateFn) -> Iterator[None]:
    with state(current_file=Path("my_example_target.py")):
        yield


@pytest.fixture()
def example_file_ir_a(make_root_context: MakeRootContextFn) -> FileIr:
    fn_a = Func(name="fn_a", interface=CallInterface())
    fn_b = Func(name="fn_b", interface=CallInterface())
    cls_a = Class(name="ClassA", interface=CallInterface(args=("self", "arg")))

    math_dot_sin = Import("sin", "math.sin")

    fn_a_ir = {
        "sets": {Name("a"), Name("b.attr", "b")},
        "gets": {Name("c")},
        "dels": set(),
        "calls": set(),
    }
    fn_b_ir = {
        "sets": set(),
        "gets": set(),
        "dels": {Name("a")},
        "calls": set(),
    }
    cls_a_ir = {
        "sets": {Name("self.field", "self")},
        "gets": {Name("arg.attr", "arg")},
        "dels": set(),
        "calls": set(),
    }

    file_ir = FileIr(
        context=make_root_context(
            symbols=[
                math_dot_sin,
                fn_a,
                fn_b,
                cls_a,
            ],
            include_root_symbols=False,
        ),
        file_ir={
            fn_a: fn_a_ir,
            fn_b: fn_b_ir,
            cls_a: cls_a_ir,
        },
    )

    return file_ir


@pytest.fixture()
def example_environment_a(example_file_ir_a: FileIr) -> IrEnvironment:
    return IrEnvironment(target_ir=example_file_ir_a, import_irs={})


@pytest.fixture()
def example_file_ir_b(make_root_context: MakeRootContextFn) -> FileIr:
    fn_a = Func(name="fn_a", interface=CallInterface(args=("a",)))
    fn_b = Func(name="fn_b", interface=CallInterface(args=("b",)))
    fn_c = Func(name="fn_c", interface=CallInterface(args=("c",)))
    fn_d = Func(name="fn_d", interface=CallInterface(args=("d",)))

    fn_b_call = Call(name="fn_b", args=CallArguments(args=("a",)), target=fn_b)
    fn_c_call = Call(name="fn_c", args=CallArguments(args=("b",)), target=fn_c)
    fn_d_call = Call(name="fn_d", args=CallArguments(args=("b",)), target=fn_d)

    fn_a_ir = {
        "sets": {Name("a")},
        "gets": set(),
        "dels": set(),
        "calls": {fn_b_call},
    }
    fn_b_ir = {
        "sets": {Name("b")},
        "gets": set(),
        "dels": set(),
        "calls": {fn_c_call, fn_d_call},
    }
    fn_c_ir = {
        "sets": {Name("c")},
        "gets": set(),
        "dels": set(),
        "calls": set(),
    }
    fn_d_ir = {
        "sets": {Name("d")},
        "gets": set(),
        "dels": set(),
        "calls": set(),
    }

    file_ir = FileIr(
        context=make_root_context(
            symbols=[
                fn_a,
                fn_b,
                fn_c,
                fn_d,
            ],
            include_root_symbols=False,
        ),
        file_ir={
            fn_a: fn_a_ir,
            fn_b: fn_b_ir,
            fn_c: fn_c_ir,
            fn_d: fn_d_ir,
        },
    )

    return file_ir


@pytest.fixture()
def example_environment_b(example_file_ir_b: FileIr) -> IrEnvironment:
    return IrEnvironment(target_ir=example_file_ir_b, import_irs={})


@pytest.fixture()
def example_file_ir_c(make_root_context: MakeRootContextFn) -> FileIr:
    fn_a = Func(name="fn_a", interface=CallInterface(args=("a",)))
    fn_b = Func(name="fn_b", interface=CallInterface(args=("b",)))

    fn_a_call = Call(name="fn_b", args=CallArguments(args=("a",)), target=fn_b)
    some_undefined_func_call = Call(
        name="some_undefined_func",
        args=CallArguments(args=()),
    )
    some_other_undefined_func_call = Call(
        name="some_other_undefined_func",
        args=CallArguments(args=()),
    )

    fn_a_ir = {
        "sets": {Name("a")},
        "gets": set(),
        "dels": set(),
        "calls": {
            fn_a_call,
            some_undefined_func_call,
        },
    }
    fn_b_ir = {
        "sets": {Name("b")},
        "gets": set(),
        "dels": set(),
        "calls": {
            some_other_undefined_func_call,
        },
    }

    file_ir = FileIr(
        context=make_root_context(
            symbols=[
                fn_a,
                fn_b,
            ],
            include_root_symbols=False,
        ),
        file_ir={
            fn_a: fn_a_ir,
            fn_b: fn_b_ir,
        },
    )

    return file_ir


@pytest.fixture()
def example_environment_c(example_file_ir_c: FileIr) -> IrEnvironment:
    return IrEnvironment(target_ir=example_file_ir_c, import_irs={})


@pytest.fixture()
def example_file_ir_d(make_root_context: MakeRootContextFn) -> FileIr:
    fn_a = Func(name="fn_a", interface=CallInterface(args=("a",)))
    os_path_join_call = Call("os.path.join", args=CallArguments(args=()))
    math_max_call = Call("math.max", args=CallArguments(args=()))

    fn_a_ir = {
        "sets": {Name("a")},
        "gets": set(),
        "dels": set(),
        "calls": {
            os_path_join_call,
            math_max_call,
        },
    }

    file_ir = FileIr(
        context=make_root_context(
            symbols=[
                fn_a,
            ],
            include_root_symbols=False,
        ),
        file_ir={
            fn_a: fn_a_ir,
        },
    )

    return file_ir


@pytest.fixture()
def example_environment_d(example_file_ir_d: FileIr) -> IrEnvironment:
    return IrEnvironment(target_ir=example_file_ir_d, import_irs={})


@pytest.fixture()
def example_file_ir_e(make_root_context: MakeRootContextFn) -> FileIr:
    fn_a = Func(name="fn_a", interface=CallInterface(args=("a",)))
    max_call = Call("max", args=CallArguments(args=()))
    enumerate_call = Call("enumerate", args=CallArguments(args=()))

    fn_a_ir = {
        "sets": {Name("a")},
        "gets": set(),
        "dels": set(),
        "calls": {
            max_call,
            enumerate_call,
        },
    }

    file_ir = FileIr(
        context=make_root_context(
            symbols=[
                fn_a,
            ],
            include_root_symbols=False,
        ),
        file_ir={
            fn_a: fn_a_ir,
        },
    )

    return file_ir


@pytest.fixture()
def example_environment_e(example_file_ir_e: FileIr) -> IrEnvironment:
    return IrEnvironment(target_ir=example_file_ir_e, import_irs={})


@pytest.fixture()
def example_file_ir_f(make_root_context: MakeRootContextFn) -> FileIr:
    fn_a = Func(name="fn_a", interface=CallInterface(args=("a",)))
    fn_b = Func(name="fn_b", interface=CallInterface(args=("b",)))
    fn_c = Func(name="fn_c", interface=CallInterface(args=("c",)))

    fn_a_ir = {
        "sets": {Name("a")},
        "gets": set(),
        "dels": set(),
        "calls": {
            Call(name="fn_a", args=CallArguments(args=("a",)), target=fn_a),
            Call(name="fn_b", args=CallArguments(args=("a",)), target=fn_b),
        },
    }
    fn_b_ir = {
        "sets": {Name("b")},
        "gets": set(),
        "dels": set(),
        "calls": {
            Call(name="fn_a", args=CallArguments(args=("b",)), target=fn_a),
            Call(name="fn_c", args=CallArguments(args=("b",)), target=fn_c),
        },
    }
    fn_c_ir = {
        "sets": {Name("b")},
        "gets": set(),
        "dels": set(),
        "calls": {
            Call(name="fn_b", args=CallArguments(args=("c",)), target=fn_b),
        },
    }

    file_ir = FileIr(
        context=make_root_context(
            symbols=[
                fn_a,
                fn_b,
                fn_c,
            ],
            include_root_symbols=False,
        ),
        file_ir={
            fn_a: fn_a_ir,
            fn_b: fn_b_ir,
            fn_c: fn_c_ir,
        },
    )

    return file_ir


@pytest.fixture()
def example_environment_f(example_file_ir_f: FileIr) -> IrEnvironment:
    return IrEnvironment(target_ir=example_file_ir_f, import_irs={})



from __future__ import annotations

from .types import (
    AstComprehensions,
    AstFunctionDef,
    AstFunctionDefOrLambda,
    AstLiterals,
)

__all__ = [
    "AstComprehensions",
    "AstFunctionDef",
    "AstFunctionDefOrLambda",
    "AstLiterals",
]



from __future__ import annotations

from collections import deque
from typing import TYPE_CHECKING

from rattr.models.results import FileResults
from rattr.results import (
    IrCallTreeNode,
    IrEnvironment,
    IrTarget,
    construct_call_swaps,
    find_call_target_and_ir,
    unbind_ir_with_call_swaps,
)

if TYPE_CHECKING:
    from rattr.analyser.types import ImportIrs
    from rattr.models.ir import FileIr, FunctionIr
    from rattr.models.symbol import Call


def generate_results_from_ir(
    *,
    target_ir: FileIr,
    import_irs: ImportIrs,
) -> FileResults:
    results = FileResults()
    environment = IrEnvironment(target_ir=target_ir, import_irs=import_irs)

    for symbol, ir in target_ir.items():
        target = IrTarget(symbol=symbol, ir=ir)

        ir_call_tree = make_target_ir_call_tree(target, environment=environment)
        simplified = destructively_simplify_ir_call_tree(ir_call_tree)

        results[symbol.id] = {
            "gets": {s.id for s in simplified["gets"]},
            "sets": {s.id for s in simplified["sets"]},
            "dels": {s.id for s in simplified["dels"]},
            "calls": {s.name_of_call for s in simplified["calls"]},
        }

    return results


def make_target_ir_call_tree(
    target: IrTarget,
    *,
    environment: IrEnvironment,
) -> IrCallTreeNode:
    """Return the constructed IR call tree stemming from the given target.

    Implementation:
        BFS the call tree from the root, constructing the nodes and children along the
        way.

    Infinite tree depth on recursion:
        A call tree may be of infinite depth due to direct or indirect recursion.
        However, if any call that has been seen before with the same parameters can be
        ignored then we can know that the finite tree with the recursive calls visited
        only once is equivalent. This assumption is proven below.

    Recursion removal proof:

        Given that:
            R(A) is the final result of a A
            for any target A

            r(A, P) is the partial result of A when called with params P
            for any target A and any parameter spec P

            R_a ∪ R_b = {k: R_a[k] ∪ R_b[k] | k ∈ {"gets", "sets", "dels", "calls"}}
            for any results R_a, R_b

            R_a ∪ R_a = R_a by the definition of ∪ over results
            for any results R_a

        Direct recursion:
            A -> A

            R(A)    = r(A, P_)
            .       ∪ r(A, P_A) ∪ r(A, P_A) ∪ ... ∪ r(A, P_A)       (by definition)
            R(A)    = r(A)
            .       ∪ r(A, P_A)                                     (idempotent)

            where P_ is the initial call's parameter spec
            and   P_A is the parameter spec for the recursive call to A made in A

        Indirect recursion:
            A -> B -> ... -> Z -> A

            R(A)    = [r(A, P_)  ∪ r(B, P_A) ∪ ... ∪ r(Z, P_Y)]
            .       ∪ [r(A, P_Z) ∪ r(B, P_A) ∪ ... ∪ r(Z, P_Y)]
            .       ...
            .       ∪ [r(A, P_Z) ∪ r(B, P_A) ∪ ... ∪ r(Z, P_Y)]     (by definition)

            R(A)    = r(A, P_)
                    ∪ [r(A, P_Z) ∪ r(B, P_A) ∪ ... ∪ r(Z, P_Y)]     (idempotent)

            where P_ is the initial call's parameter spec
            and   P_N is the parameter spec for a call as seen in target N

        There are more complex cases (direct recursion where A calls out to other
        functions, the same for each step in the case of indirect recursion, etc),
        however, those trivially hold given the above and are cumbersome to express.
    """
    root = IrCallTreeNode.new(target=target, call=None)

    queue = deque([root])
    seen: set[Call] = set()

    while queue:
        node = queue.popleft()

        for call in node.edges_out:
            if call.symbol in seen:
                continue

            call_target = find_call_target_and_ir(call, environment=environment)

            if call_target is None:
                continue

            child = IrCallTreeNode.new(target=call_target, call=call)

            node.children.append(child)
            queue.append(child)

            seen.add(call.symbol)

    return root


def destructively_simplify_ir_call_tree(root: IrCallTreeNode) -> FunctionIr:
    queue = post_order_traversal_queue(root)

    while queue:
        node = queue.popleft()

        if not node.children:
            # Leaves are already simplified
            continue

        for child in node.children:
            swaps = construct_call_swaps(child.target.symbol, child.edge_in.symbol)
            unbound = unbind_ir_with_call_swaps(child.target.ir, swaps)

            node.target.ir["sets"] |= unbound["sets"]
            node.target.ir["gets"] |= unbound["gets"]
            node.target.ir["dels"] |= unbound["dels"]

    return root.target.ir


def post_order_traversal_queue(root: IrCallTreeNode) -> deque[IrCallTreeNode]:
    queue = deque([root])
    pre_order_queue = list()

    while queue:
        node = queue.popleft()
        for child in node.children:
            queue.append(child)
        pre_order_queue.append(node)

    return deque(pre_order_queue[::-1])



from __future__ import annotations

from typing import TYPE_CHECKING

from rattr.models.ir import FileIr
from rattr.models.util._serialisation_helpers import make_json_converter
from rattr.models.util._types import FileName, ImportIrs, OutputIrs

if TYPE_CHECKING:
    from typing import Any, TypeVar

    T = TypeVar("T")


__json_converter = make_json_converter()


def serialise(model: Any, **kwargs) -> str:
    return __json_converter.dumps(model, **kwargs)


def deserialise(json: str, *, type: type[T], **kwargs) -> T:
    return __json_converter.loads(json, cl=type, **kwargs)


def serialise_irs(
    *,
    target_name: FileName,
    target_ir: FileIr,
    import_irs: ImportIrs,
) -> str:
    return serialise(
        OutputIrs(
            import_irs=import_irs,
            target_ir={"filename": target_name, "ir": target_ir},
        ),
        indent=4,
    )



from __future__ import annotations

import ast
from collections.abc import MutableMapping
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING, Union

import attrs
from attrs import field

from rattr import error
from rattr.ast.types import Identifier
from rattr.ast.util import unravel_names
from rattr.config.state import enter_file
from rattr.config.util import get_current_file
from rattr.models.context._symbol_table import SymbolTable
from rattr.models.context._util import (
    is_call_to_call_result,
    is_call_to_literal,
    is_call_to_member_of_module_import,
    is_call_to_method,
    is_call_to_method_on_py_type,
    is_call_to_subscript_item,
)
from rattr.models.symbol._symbol import AnyCallInterface, CallInterface, Symbol
from rattr.models.symbol._symbols import Class, Func, Import, Name
from rattr.models.symbol.util import (
    with_call_brackets,
    without_call_brackets,
)
from rattr.module_locator.util import (
    derive_module_name_from_path,
    derive_module_names_right,
)

if TYPE_CHECKING:
    from collections.abc import Container, Iterable, Iterator
    from typing import Protocol, TypeVar

    from rattr.models.symbol._types import CallableSymbol

    class ContainerWithContext(Protocol):
        context: Context

    SymbolType = TypeVar("SymbolType", bound=Symbol)


@contextmanager
def new_context(container: ContainerWithContext) -> Iterator[None]:
    """Set the container context to a new child in the context block, pop on exit."""
    parent = container.context
    container.context = Context(parent)
    yield
    container.context = parent


@attrs.mutable
class Context(MutableMapping[Identifier, Symbol]):
    parent: Union[Context, None]
    symbol_table: SymbolTable = field(factory=SymbolTable, kw_only=True)

    _file: Path = field(
        alias="file",
        factory=get_current_file,
        kw_only=True,
        hash=False,
        eq=False,
    )

    @property
    def file(self) -> Path:
        return self._file

    @property
    def root(self) -> Context:
        if self.parent is None:
            return self

        return self.parent.root

    @property
    def is_init_file(self) -> bool:
        return self.file.name == "__init__.py"

    @property
    def modulename(self) -> str:
        return derive_module_name_from_path(self.file)

    @property
    def declared_symbols(self) -> set[Symbol]:
        """The symbols declared in this scope (excl. symbols in ancestor scopes)."""
        return set(self.symbol_table.symbols)

    @property
    def all_symbols(self) -> set[Symbol]:
        """The symbols present in this scope (incl. symbols in ancestor scopes)."""
        if self.parent is None:
            return self.declared_symbols

        return self.declared_symbols | self.parent.all_symbols

    @property
    def declared_names(self) -> set[Identifier]:
        return set(self.symbol_table.names)

    @property
    def all_names(self) -> set[Identifier]:
        if self.parent is None:
            return self.declared_names

        return self.declared_names | self.parent.all_names

    def add(
        self,
        symbol_or_symbols: Symbol | Iterable[Symbol],
        *,
        is_argument: bool = False,
    ) -> None:
        """Add the given symbol(s) to the context.

        If the symbol is already present in the context, or an ancestor, then
        it will not be re-added.

        If the symbol `is_argument`, then the symbol will always be added
        regardless of whether-or-not it is declared in an ancestor.
        """
        if isinstance(symbol_or_symbols, Symbol):
            symbols = [symbol_or_symbols]
        else:
            symbols = symbol_or_symbols

        for symbol in (s for s in symbols if s.name not in self or is_argument):
            self[symbol.name] = symbol

    def remove(self, id_or_ids: Identifier | Iterable[Identifier]) -> None:
        """Remove the given id(s) from the context.

        If the symbol is declared in this context it will be removed, likewise if it is
        declared in a ancestor it will also be removed.
        """
        if isinstance(id_or_ids, Identifier):
            ids = [id_or_ids]
        else:
            ids = id_or_ids

        for id in ids:
            self.symbol_table.pop(id)

    def delete(self, id_or_ids: Identifier | Iterable[Identifier]) -> None:
        """Alias to remove."""
        self.remove(id_or_ids)

    def get_call_target(
        self,
        callee: Identifier,
        culprit: ast.Call | ast.FunctionDef,
        *,
        warn: bool = True,
    ) -> CallableSymbol | None:
        """Return the target of the given call as resolved within this context.

        Args:
            callee (Identifier): The callee name.
            culprit (ast.Call | ast.FunctionDef): \
                The ast.Call node, used as the error culprit. This can be an \
                ast.FunctionDef if it comes from a `rattr_results` decorator for \
                example.
            warn (bool, optional): If `True` give warnings. Defaults to True.

        Returns:
            CallableSymbol | None: The callable symbol.
        """
        canonical = with_call_brackets(callee)
        _error = f"unable to resolve call to {canonical!r}, {{reason}}"

        name = without_call_brackets(callee).replace("*", "")
        (lhs_name, *_) = name.split(".")

        if is_call_to_literal(name):
            if warn:
                error.info(
                    _error.format(reason="target lhs is a literal"),
                    culprit=culprit,
                )
            return None

        if is_call_to_subscript_item(name):
            if warn:
                if "." in name:
                    error.info(
                        _error.format(reason="target lhs is run-time dependent"),
                        culprit=culprit,
                    )
                else:
                    error.error(
                        _error.format(reason="target is run-time dependent"),
                        culprit=culprit,
                    )
            return None

        target = self.get(name)
        lhs_target = self.get(lhs_name)

        if is_call_to_method(target, name, lhs_target, lhs_name):
            if warn and not is_call_to_method_on_py_type(name):
                error.info(
                    _error.format(reason="target is a method"),
                    culprit=culprit,
                )
            return target

        # Check for calls to members of imported modules, i.e. the second case below:
        #   `from math import pi`   ->  pi is in context, already resolved above
        #   `import math`           ->  `math.pi` is not explicitly in context
        if is_call_to_member_of_module_import(name, target):
            target = self._get_target_in_imported_module(name)

        _target_is_callable = target is not None and target.is_callable

        # Give warnings for unresolvable targets
        if target is None:
            if warn:
                error.warning(
                    _error.format(reason="target is undefined"),
                    culprit=culprit,
                )
            return None

        if is_call_to_call_result(culprit):
            if warn:
                error.error(
                    _error.format(reason="target is a call on a call"),
                    culprit=culprit,
                )
            return target

        # Give warnings for targets likely to be resolved incorrectly
        # TODO On method support, upgrade info to warning
        if not _target_is_callable:
            if warn:
                if target is not None:
                    if "." not in name and self.declares(target.name):
                        error.error(
                            _error.format(
                                reason="target is likely a procedural parameter",
                            ),
                            culprit=culprit,
                        )
                    elif "." in target.name:
                        error.info(
                            _error.format(reason="target is a method"),
                            culprit=culprit,
                        )
                    else:
                        error.error(
                            _error.format(reason="target is not callable"),
                            culprit=culprit,
                        )
                else:
                    error.error(
                        _error.format(reason="target is not callable"),
                        culprit=culprit,
                    )
            return target

        return target

    # TODO Note to self: I don't like this method name, change it!
    def declares(self, id: Identifier) -> bool:
        """Return `True` if the id was defined in this context, not a parent."""
        return id in self.symbol_table

    def get_starred_imports(
        self,
        *,
        seen_by_origin: Container[Path] | None = None,
    ) -> list[Import]:
        """Return the starred imports in the current context, skipping previously seen.

        Args:
            seen (Container[str] | None, optional):
                The previously seen starred imports by the module spec's origin.
                Defaults to None.

        Returns:
            list[Import]: The starred imports.
        """
        return [
            symbol
            for symbol in self.declared_symbols
            if isinstance(symbol, Import) and symbol.name == "*"
            if symbol.origin not in seen_by_origin or ()
        ]

    # ================================================================================ #
    # Registration helpers
    # ================================================================================ #

    def add_identifiers_to_context(self, assignment: ast.expr) -> None:
        self.add(Name(name, token=assignment) for name in unravel_names(assignment))

    def remove_identifiers_from_context(self, assignment: ast.expr) -> None:
        self.remove(unravel_names(assignment))

    def add_arguments_to_context(
        self,
        arguments: ast.arguments,
        *,
        token: ast.AST,
    ) -> None:
        self.add(
            Name(arg, token=token)
            for arg in CallInterface.from_arguments(arguments).all
        )

    # ================================================================================ #
    # Syntactic sugar methods
    # ================================================================================ #

    def expand_starred_imports(self) -> Context:
        """Recursively follow starred imports and add the discovered names to the scope.

        This is an in-place operation which returns self.
        """
        from rattr.models.context._root_context import compile_root_context

        seen: set[Path] = set()
        queue = self.get_starred_imports(seen_by_origin=seen)

        # BFS the unseen starred imports
        for starred in queue:
            if starred.origin is None:
                error.error(
                    f"unable to resolve import {starred.name!r} while expanding "
                    f"{starred.code()!r}",
                    culprit=starred,
                )
                continue

            if starred.origin in seen:
                continue

            # Visit node
            with enter_file(starred.origin):
                starred_ast = ast.parse(starred.origin.read_text())
                starred_context = compile_root_context(starred_ast)

            # Progress breadth-first search queue
            seen.add(starred.origin)
            queue += starred_context.get_starred_imports(seen_by_origin=seen)

            # Add the resolved names to this context
            for symbol in starred_context.declared_symbols:
                self.add(
                    Import(
                        name=symbol.name,
                        qualified_name=f"{starred.qualified_name}.{symbol.name}",
                        location=starred.location,
                        interface=symbol.interface,
                    )
                )

        return self

    def get_class_or_error(self, name: Identifier) -> Class:
        cls = self.get(name)

        if cls is None or not isinstance(cls, Class):
            raise KeyError(f"class {name!r} is not defined in the current context")

        return cls

    def get_func_or_error(self, name: Identifier) -> Func:
        func = self.get(name)

        if func is None or not isinstance(func, Func):
            raise KeyError(f"function {name!r} is not defined in the current context")

        return func

    # ================================================================================ #
    # Private helper methods
    # ================================================================================ #

    def _get_target_in_imported_module(self, name: Identifier) -> Symbol | None:
        modules = [self.get(m) for m in derive_module_names_right(name)]
        module = next((m for m in modules if m is not None), None)

        if not isinstance(module, Import):
            return None

        local_name = name.replace(f"{module.name}.", "")

        return Import(
            name=local_name,
            qualified_name=f"{module.qualified_name}.{local_name}",
            location=module.location,
            interface=AnyCallInterface(),
        )

    # ================================================================================ #
    # Mutable mapping abstract methods and mixin-overrides
    # ================================================================================ #

    def __getitem__(self, __key: Identifier) -> Symbol:
        if __key in self.symbol_table:
            return self.symbol_table.__getitem__(__key)

        if self.parent is None:
            raise KeyError(__key)

        return self.parent.__getitem__(__key)

    def __setitem__(self, __key: Identifier, __value: Symbol) -> None:
        _error = f"symbol key and id do not match: {__key!r} != {__value.id!r}"

        if __key == "*" and not __value.id.endswith(".*"):
            raise ValueError(_error)

        if __key != "*" and __key != __value.id:
            raise ValueError(_error)

        return self.symbol_table.add(__value)

    def __delitem__(self, __key: Identifier) -> None:
        if __key in self.symbol_table:
            return self.symbol_table.__delitem__(__key)

        if self.parent is None:
            raise KeyError(__key)

        return self.parent.__delitem__(__key)

    def __iter__(self) -> Iterator[Identifier]:
        # Get the chain of ancestors from here to root
        contexts: list[Context] = [ctx := self]

        while (ctx := ctx.parent) is not None:
            contexts.append(ctx)

        # Yield the symbols from each context, from the root context to this context,
        # in order of declaration. As dicts are insert-ordered, and symbol tables are
        # dicts with symbols inserted in declaration order, order holds if we order the
        # contexts from root to current.
        for context in reversed(contexts):
            yield from context.symbol_table.__iter__()

    def __len__(self) -> int:
        return self.all_symbols.__len__()

    def clear(self) -> None:
        raise TypeError("a context is mutable but not re-usable")

    def update(self) -> None:
        raise TypeError("a context is mutable but not re-usable")



from __future__ import annotations

from typing import NamedTuple


class CallInterfaceArgs(NamedTuple):
    posonlyargs: list[str]
    args: list[str]
    vararg: str | None
    kwonlyargs: list[str]
    kwarg: str | None



from __future__ import annotations

import sys
from functools import partial
from pathlib import Path

import pytest

short_relative_path = Path("relative_dir") / "file.txt"
long_relative_path = (
    Path("relative_dir")
    / "with"
    / "some"
    / "deeply"
    / "nested"
    / "content"
    / "inside.it"
)


short_absolute_path_in_home = Path.home().resolve() / "rattr" / "target.py"
long_absolute_path_in_home = (
    Path.home().resolve()
    / "some"
    / "deeply"
    / "nested"
    / "path"
    / "to"
    / "some"
    / "rattr"
    / "target.py"
)


short_absolute_path_not_in_home = Path("/tmp") / "my_temp.file"
long_absolute_path_not_in_home = (
    Path("/tmp") / "a" / "b" / "c" / "dee" / "eff" / "jee" / "my_temp.file"
)

short_absolute_path_not_in_home_windows = Path("C:/") / "Program Files" / "thefile.win"
long_absolute_path_not_in_home_windows = (
    Path("C:/")
    / "Program Files"
    / "Rattr"
    / "Config"
    / "Made Up"
    / "Path"
    / "Here"
    / "thefile.win"
)


@pytest.fixture
def reset(state):
    return partial(
        state,
        badness_from_simplification=0,
        badness_from_target_file=0,
        badness_from_imports=0,
    )


@pytest.fixture
def assert_badness(config):
    def _make_assertion(simplification: int, target: int, imports: int):
        assert config.state.badness_from_simplification == simplification
        assert config.state.badness_from_target_file == target
        assert config.state.badness_from_imports == imports

    return _make_assertion


@pytest.fixture
def badness(state):
    def _set_badness(simplification: int, target: int, imports: int):
        return state(
            badness_from_simplification=simplification,
            badness_from_target_file=target,
            badness_from_imports=imports,
        )

    return _set_badness


class TestState:
    def test_badness_vs_full_badness(self, badness, config):
        # For now at least these are expected to differ, thus there should be a test
        # for regression purposes at the very least.
        with badness(simplification=3, target=5, imports=7):
            assert config.state.badness == 8
            assert config.state.full_badness == 15


class TestConfig:
    def test_increment_badness(self, assert_badness, config, state, reset):
        with reset():
            assert_badness(simplification=0, target=0, imports=0)

            with state(current_file=None):
                config.increment_badness(5)

            assert_badness(simplification=5, target=0, imports=0)

        with reset():
            assert_badness(simplification=0, target=0, imports=0)

            with state(current_file=Path("target.py")):
                config.increment_badness(5)

            assert_badness(simplification=0, target=5, imports=0)

        with reset():
            assert_badness(simplification=0, target=0, imports=0)

            with state(current_file="not none but not the target"):
                config.increment_badness(5)

            assert_badness(simplification=0, target=0, imports=5)

    def test_is_within_badness_threshold_is_strict(self, config, arguments, badness):
        with arguments(is_strict=True):
            with badness(simplification=0, target=0, imports=0):
                assert config.is_within_badness_threshold

            with badness(simplification=1, target=0, imports=0):
                assert not config.is_within_badness_threshold

            with badness(simplification=0, target=1, imports=0):
                assert not config.is_within_badness_threshold

            with badness(simplification=0, target=0, imports=1):
                assert config.is_within_badness_threshold

    def test_is_within_badness_threshold_infinite_threshold(
        self, config, arguments, badness
    ):
        with arguments(is_strict=False, threshold=0):
            with badness(simplification=0, target=0, imports=0):
                assert config.is_within_badness_threshold

            with badness(simplification=1_000, target=0, imports=0):
                assert config.is_within_badness_threshold

            with badness(simplification=0, target=1_000, imports=0):
                assert config.is_within_badness_threshold

            with badness(simplification=0, target=0, imports=1_000):
                assert config.is_within_badness_threshold

    def test_is_within_badness_threshold_ordinary(self, config, arguments, badness):
        with arguments(is_strict=False, threshold=500):
            with badness(simplification=0, target=0, imports=0):
                assert config.is_within_badness_threshold

            # Simplification badness is part of badness
            with badness(simplification=499, target=0, imports=0):
                assert config.is_within_badness_threshold
            with badness(simplification=500, target=0, imports=0):
                assert config.is_within_badness_threshold
            with badness(simplification=501, target=0, imports=0):
                assert not config.is_within_badness_threshold

            # Target file badness is part of badness
            with badness(simplification=0, target=499, imports=0):
                assert config.is_within_badness_threshold
            with badness(simplification=0, target=500, imports=0):
                assert config.is_within_badness_threshold
            with badness(simplification=0, target=501, imports=0):
                assert not config.is_within_badness_threshold

            # Imported badness is *not* part of badness (only full_badness)
            with badness(simplification=0, target=0, imports=499):
                assert config.is_within_badness_threshold
            with badness(simplification=0, target=0, imports=500):
                assert config.is_within_badness_threshold
            with badness(simplification=0, target=0, imports=501):
                assert config.is_within_badness_threshold

            with badness(simplification=250, target=250, imports=101):
                assert config.is_within_badness_threshold
            with badness(simplification=251, target=250, imports=101):
                assert not config.is_within_badness_threshold

    def test_formatted_current_file_path(self, config, arguments, state):
        file = Path.home() / "this" / "is" / "a" / "path" / "to" / "a.file"

        with arguments(collapse_home=True, truncate_deep_paths=True):
            with state(current_file=None):
                assert config.formatted_current_file_path is None

            with state(current_file=file):
                assert config.formatted_current_file_path == "~/.../path/to/a.file"

    def test_formatted_target_path(self, config, arguments):
        file = Path.home() / "this" / "is" / "a" / "path" / "to" / "a.file"

        with arguments(collapse_home=True, truncate_deep_paths=True):
            with arguments(target=None):
                assert config.formatted_target_path is None

            with arguments(target=file):
                assert config.formatted_target_path == "~/.../path/to/a.file"


class TestGetFormattedPath:
    def test_path_is_none(self, config):
        assert config.get_formatted_path(None) is None

    @pytest.mark.parametrize(
        "path,expected",
        [
            (short_relative_path, short_relative_path.as_posix()),
            (long_relative_path, long_relative_path.as_posix()),
            (short_absolute_path_in_home, short_absolute_path_in_home.as_posix()),
            (long_absolute_path_in_home, long_absolute_path_in_home.as_posix()),
            (
                short_absolute_path_not_in_home,
                short_absolute_path_not_in_home.as_posix(),
            ),
            (long_absolute_path_not_in_home, long_absolute_path_not_in_home.as_posix()),
        ],
        ids=[
            "short_relative_path",
            "long_relative_path",
            "short_absolute_path_in_home",
            "long_absolute_path_in_home",
            "short_absolute_path_not_in_home",
            "long_absolute_path_not_in_home",
        ],
    )
    def test_raw_path(self, config, arguments, path, expected):
        with arguments(collapse_home=False, truncate_deep_paths=False):
            assert config.get_formatted_path(path) == expected

    @pytest.mark.parametrize(
        "path,expected",
        [
            (short_relative_path, short_relative_path.as_posix()),
            (long_relative_path, long_relative_path.as_posix()),
            (short_absolute_path_in_home, "~/rattr/target.py"),
            (
                long_absolute_path_in_home,
                "~/some/deeply/nested/path/to/some/rattr/target.py",
            ),
            (
                short_absolute_path_not_in_home,
                short_absolute_path_not_in_home.as_posix(),
            ),
            (long_absolute_path_not_in_home, long_absolute_path_not_in_home.as_posix()),
        ],
        ids=[
            "short_relative_path",
            "long_relative_path",
            "short_absolute_path_in_home",
            "long_absolute_path_in_home",
            "short_absolute_path_not_in_home",
            "long_absolute_path_not_in_home",
        ],
    )
    def test_collapsed_home(self, config, arguments, path, expected):
        with arguments(collapse_home=True, truncate_deep_paths=False):
            assert config.get_formatted_path(path) == expected

    @pytest.mark.parametrize(
        "path,expected",
        [
            (short_relative_path, short_relative_path.as_posix()),
            (long_relative_path, "relative_dir/.../nested/content/inside.it"),
            (short_absolute_path_in_home, short_absolute_path_in_home.as_posix()),
            (long_absolute_path_in_home, "/.../some/rattr/target.py"),
            (
                short_absolute_path_not_in_home,
                short_absolute_path_not_in_home.as_posix(),
            ),
            (long_absolute_path_not_in_home, "/.../eff/jee/my_temp.file"),
        ],
        ids=[
            "short_relative_path",
            "long_relative_path",
            "short_absolute_path_in_home",
            "long_absolute_path_in_home",
            "short_absolute_path_not_in_home",
            "long_absolute_path_not_in_home",
        ],
    )
    def test_truncate_deep_paths(self, config, arguments, path, expected):
        if path == long_absolute_path_in_home and sys.platform == "win32":
            expected = f"C:{expected}"

        with arguments(collapse_home=False, truncate_deep_paths=True):
            assert config.get_formatted_path(path) == expected

    @pytest.mark.parametrize(
        "path,expected",
        [
            (short_relative_path, short_relative_path.as_posix()),
            (long_relative_path, "relative_dir/.../nested/content/inside.it"),
            (short_absolute_path_in_home, "~/rattr/target.py"),
            (long_absolute_path_in_home, "~/.../some/rattr/target.py"),
            (
                short_absolute_path_not_in_home,
                short_absolute_path_not_in_home.as_posix(),
            ),
            (long_absolute_path_not_in_home, "/.../eff/jee/my_temp.file"),
        ],
        ids=[
            "short_relative_path",
            "long_relative_path",
            "short_absolute_path_in_home",
            "long_absolute_path_in_home",
            "short_absolute_path_not_in_home",
            "long_absolute_path_not_in_home",
        ],
    )
    def test_fully_formatted(self, config, arguments, path, expected):
        with arguments(collapse_home=True, truncate_deep_paths=True):
            assert config.get_formatted_path(path) == expected

    @pytest.mark.windows
    def test_windows(self, config, arguments):
        with arguments(collapse_home=False, truncate_deep_paths=False):
            assert (
                config.get_formatted_path(short_absolute_path_not_in_home_windows)
                == "C:/Program Files/thefile.win"
            )
            assert (
                config.get_formatted_path(long_absolute_path_not_in_home_windows)
                == "C:/Program Files/Rattr/Config/Made Up/Path/Here/thefile.win"
            )

        with arguments(collapse_home=False, truncate_deep_paths=True):
            assert (
                config.get_formatted_path(short_absolute_path_not_in_home_windows)
                == "C:/Program Files/thefile.win"
            )
            assert (
                config.get_formatted_path(long_absolute_path_not_in_home_windows)
                == "C:/.../Path/Here/thefile.win"
            )

        with arguments(collapse_home=True, truncate_deep_paths=False):
            assert (
                config.get_formatted_path(short_absolute_path_not_in_home_windows)
                == "C:/Program Files/thefile.win"
            )
            assert (
                config.get_formatted_path(long_absolute_path_not_in_home_windows)
                == "C:/Program Files/Rattr/Config/Made Up/Path/Here/thefile.win"
            )

        with arguments(collapse_home=True, truncate_deep_paths=True):
            assert (
                config.get_formatted_path(short_absolute_path_not_in_home_windows)
                == "C:/Program Files/thefile.win"
            )
            assert (
                config.get_formatted_path(long_absolute_path_not_in_home_windows)
                == "C:/.../Path/Here/thefile.win"
            )



from __future__ import annotations

import ast
from typing import TYPE_CHECKING

import attrs
from cattrs.gen import make_dict_structure_fn, make_dict_unstructure_fn, override
from cattrs.preconf.json import make_converter

from rattr.models.context import Context, SymbolTable
from rattr.models.ir import FileIr, FunctionIr
from rattr.models.results import FileResults, FunctionName, FunctionResults
from rattr.models.symbol import (
    AnyCallInterface,
    Builtin,
    Call,
    CallInterface,
    Class,
    Func,
    Import,
    Location,
    Name,
    Symbol,
)

if TYPE_CHECKING:
    from typing import Any, Final, Literal

    from cattrs.converters import Converter

    from rattr.ast.types import Identifier


__symbols: Final = (Name, Builtin, Import, Func, Class, Call)
__name_to_symbol: Final = {symbol.__name__: symbol for symbol in __symbols}


def make_json_converter() -> Converter:
    converter = make_converter()

    converter.register_structure_hook(
        Location,
        make_dict_structure_fn(
            Location,
            converter,
            _cattrs_use_alias=True,
            token=override(omit=True),
        ),
    )
    converter.register_unstructure_hook(
        Location,
        make_dict_unstructure_fn(
            Location,
            converter,
            _cattrs_use_alias=True,
            token=override(omit=True),
        ),
    )

    call_interface_converter = make_converter()
    converter.register_structure_hook(
        CallInterface,
        make_call_interface_deserialiser(call_interface_converter),
    )
    converter.register_unstructure_hook(
        CallInterface,
        make_call_interface_serialiser(call_interface_converter),
    )
    converter.register_structure_hook(
        AnyCallInterface,
        make_call_interface_deserialiser(call_interface_converter),
    )
    converter.register_unstructure_hook(
        AnyCallInterface,
        make_call_interface_serialiser(call_interface_converter),
    )

    symbol_converter = make_symbol_json_converter()
    converter.register_structure_hook(
        Symbol,
        make_symbol_deserialiser(symbol_converter),
    )
    converter.register_unstructure_hook(
        Symbol,
        make_symbol_serialiser(symbol_converter),
    )

    converter.register_structure_hook(
        SymbolTable,
        make_symbol_table_deserialiser(converter),
    )
    converter.register_unstructure_hook(
        SymbolTable,
        make_symbol_table_serialiser(converter),
    )

    converter.register_structure_hook(
        Context,
        make_dict_structure_fn(
            Context,
            converter,
            _cattrs_use_alias=True,
            _cattrs_include_init_false=True,
        ),
    )
    converter.register_unstructure_hook(
        Context,
        make_dict_unstructure_fn(
            Context,
            converter,
            _cattrs_use_alias=True,
            _cattrs_include_init_false=True,
        ),
    )

    converter.register_structure_hook(
        FileResults,
        make_file_results_deserialiser(converter),
    )
    converter.register_unstructure_hook(
        FileResults,
        make_file_results_serialiser(converter),
    )

    converter.register_structure_hook(FileIr, make_file_ir_deserialiser(converter))
    converter.register_unstructure_hook(FileIr, make_file_ir_serialiser(converter))

    return converter


def make_symbol_json_converter() -> Converter:
    converter = make_converter()

    converter.register_structure_hook(
        Location,
        make_dict_structure_fn(
            Location,
            converter,
            _cattrs_use_alias=True,
            token=override(omit=True),
        ),
    )
    converter.register_unstructure_hook(
        Location,
        make_dict_unstructure_fn(
            Location,
            converter,
            _cattrs_use_alias=True,
            token=override(omit=True),
        ),
    )

    call_interface_converter = make_converter()
    converter.register_structure_hook(
        CallInterface,
        make_call_interface_deserialiser(call_interface_converter),
    )
    converter.register_unstructure_hook(
        CallInterface,
        make_call_interface_serialiser(call_interface_converter),
    )
    converter.register_structure_hook(
        AnyCallInterface,
        make_call_interface_deserialiser(call_interface_converter),
    )
    converter.register_unstructure_hook(
        AnyCallInterface,
        make_call_interface_serialiser(call_interface_converter),
    )

    converter.register_structure_hook(ast.AST, lambda _, __: None)
    converter.register_unstructure_hook(ast.AST, lambda _: None)

    return converter


def make_symbol_serialiser(converter: Converter):
    def serialise_symbol(symbol: Symbol) -> dict[str, Any]:
        data: dict[str, Any] = converter.unstructure(symbol)
        data.pop("token", None)  # not serialisable, should be None form raw converter
        data = {"type": symbol.__class__.__name__, **data}

        if isinstance(symbol, Call) and data["target"] is not None:
            data["target"] = serialise_symbol(symbol.target)

        return data

    return serialise_symbol


def make_symbol_deserialiser(converter: Converter):
    def deserialise_symbol(data: dict[str, Any], _: type[Symbol]) -> Symbol:
        type: str = data.pop("type", None)

        if type is None:
            raise ValueError("missing type")

        if type not in __name_to_symbol.keys():
            raise ValueError(f"invalid type: {type}")

        if type == Call.__name__:
            target_data = data.pop("target", None)

            if target_data is not None:
                target = deserialise_symbol(target_data, _)
            else:
                target = None

            data["target"] = None
            symbol = converter.structure(data, __name_to_symbol[type])

            return attrs.evolve(symbol, target=target)

        return converter.structure(data, __name_to_symbol[type])

    return deserialise_symbol


def make_call_interface_serialiser(converter: Converter):
    def serialise_call_interface(
        call_interface: CallInterface,
    ) -> dict[str, Any] | Literal["any"]:
        if isinstance(call_interface, AnyCallInterface):
            return "any"
        else:
            return converter.unstructure(call_interface)

    return serialise_call_interface


def make_call_interface_deserialiser(converter: Converter):
    def deserialise_call_interface(
        data: dict[str, Any] | Literal["any"] | object,
        type: type[CallInterface],
    ) -> CallInterface:
        if data == "any":
            return AnyCallInterface()
        elif isinstance(data, dict):
            return converter.structure(data, CallInterface)
        else:
            raise ValueError(f"not a valid call interface: {data}")

    return deserialise_call_interface


def make_symbol_table_serialiser(converter: Converter):
    def serialise_symbol_table(symbol_table: SymbolTable) -> dict[str, Any]:
        return converter.unstructure(symbol_table._symbols)

    return serialise_symbol_table


def make_symbol_table_deserialiser(converter: Converter):
    def deserialise_symbol_table(
        data: dict[Identifier | object, dict[str, Any] | object],
        cls: type[SymbolTable],
    ) -> SymbolTable:
        symbol_table = cls()
        symbol_table._symbols = converter.structure(data, dict[str, Symbol])
        return symbol_table

    return deserialise_symbol_table


def make_file_results_serialiser(_: Converter):
    def serialise_file_results(file_results: FileResults) -> dict[str, FunctionResults]:
        return {
            name: {
                "gets": sorted(file_results._function_results[name]["gets"]),
                "sets": sorted(file_results._function_results[name]["sets"]),
                "dels": sorted(file_results._function_results[name]["dels"]),
                "calls": sorted(file_results._function_results[name]["calls"]),
            }
            for name in sorted(file_results._function_results.keys())
        }

    return serialise_file_results


def make_file_results_deserialiser(converter: Converter):
    def deserialise_file_results(
        data: dict[Identifier | object, dict[str, Any] | object],
        cls: type[FileResults],
    ) -> FileResults:
        return cls(converter.structure(data, dict[FunctionName, FunctionResults]))

    return deserialise_file_results


def make_file_ir_serialiser(converter: Converter):
    def serialise_file_ir(file_ir: FileIr) -> dict[str, Any]:
        return {
            "context": converter.unstructure(file_ir.context),
            "symbols": {
                symbol.id: converter.unstructure(symbol)
                for symbol in sorted(file_ir._file_ir.keys())
            },
            "function_irs": {
                symbol.id: {
                    "gets": sorted(
                        converter.unstructure(file_ir._file_ir[symbol]["gets"]),
                        key=lambda s: s["name"],
                    ),
                    "sets": sorted(
                        converter.unstructure(file_ir._file_ir[symbol]["sets"]),
                        key=lambda s: s["name"],
                    ),
                    "dels": sorted(
                        converter.unstructure(file_ir._file_ir[symbol]["dels"]),
                        key=lambda s: s["name"],
                    ),
                    "calls": sorted(
                        converter.unstructure(file_ir._file_ir[symbol]["calls"]),
                        key=lambda s: s["name"],
                    ),
                }
                for symbol in sorted(file_ir._file_ir.keys())
            },
        }

    return serialise_file_ir


def make_file_ir_deserialiser(converter: Converter):
    def deserialise_file_ir(data: dict[str, Any], cls: type[FileIr]) -> FileIr:
        context_data: dict[str, Any] = data.get("context", None)
        context = (
            converter.structure(context_data, Context)
            if context_data is not None
            else None
        )

        symbols_data: dict[Identifier, dict[str, Any]] = data.get("symbols", None)
        fn_irs_data: dict[Identifier, dict[str, Any]] = data.get("function_irs", None)

        if symbols_data is None or fn_irs_data is None:
            raise ValueError

        symbols = {
            identifier: converter.structure(symbol_data, Symbol)
            for identifier, symbol_data in symbols_data.items()
        }
        fn_irs = {
            identifier: converter.structure(ir_data, FunctionIr)
            for identifier, ir_data in fn_irs_data.items()
        }

        return cls(
            context=context,
            file_ir={
                symbol: value
                for name, symbol in symbols.items()
                if (value := fn_irs.get(name, None)) is not None
            },
        )

    return deserialise_file_ir



from __future__ import annotations

import ast
from typing import TYPE_CHECKING
from unittest import mock

import pytest

from rattr.models.context import Context, compile_root_context
from rattr.models.context._root_context import MODULE_LEVEL_DUNDER_ATTRS
from rattr.models.symbol import CallInterface, Class, Func, Import, Name
from tests.shared import compare_symbol_table_symbols, match_output

if TYPE_CHECKING:
    from rattr.models.symbol import Builtin
    from tests.shared import ArgumentsFn, MakeSymbolTableFn, ParseFn


@pytest.fixture
def root_context() -> Context:
    return compile_root_context(ast.Module(body=[]))


def test_root_context_includes_module_level_attributes(root_context: Context):
    for attr in MODULE_LEVEL_DUNDER_ATTRS:
        assert attr in root_context.symbol_table.names


def test_root_context_includes_python_builtins(root_context: Context):
    example_builtins = (
        # functions
        "print",
        "max",
        "len",
        # classes / errors
        "Exception",
        "KeyError",
        "TypeError",
    )

    for attr in example_builtins:
        assert attr in root_context.symbol_table.names


def test_root_context_python_builtins_have_correct_affect_status(root_context: Context):
    print_symbol: Builtin = root_context.get("print")
    setattr_symbol: Builtin = root_context.get("setattr")

    assert not print_symbol.has_affect
    assert setattr_symbol.has_affect


def test_root_context_import(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        import math
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Import(name="math", qualified_name="math", location=mock.ANY),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_import_as(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        import math as m
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Import(name="m", qualified_name="math", location=mock.ANY),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_import_list(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
    arguments: ArgumentsFn,
    capfd: pytest.CaptureFixture[str],
):
    module_ast = parse(
        """
        import os, math
        """
    )

    with arguments(_warning_level="all"):
        context = compile_root_context(module_ast)

    expected = make_symbol_table(
        [
            Import(name="os", qualified_name="os", location=mock.ANY),
            Import(name="math", qualified_name="math", location=mock.ANY),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)

    _, stderr = capfd.readouterr()
    assert match_output(stderr, ["do not import multiple modules on one line"])


def test_root_context_import_from(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        from os.path import isfile
        from math import sin, cos
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Import(name="isfile", qualified_name="os.path.isfile", location=mock.ANY),
            Import(name="sin", qualified_name="math.sin", location=mock.ANY),
            Import(name="cos", qualified_name="math.cos", location=mock.ANY),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_import_from_as(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        from os import path as path_utils
        from math import power as exp, cos
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Import(name="path_utils", qualified_name="os.path", location=mock.ANY),
            Import(name="exp", qualified_name="math.power", location=mock.ANY),
            Import(name="cos", qualified_name="math.cos", location=mock.ANY),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_import_from_star(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
    capfd: pytest.CaptureFixture[str],
):
    module_ast = parse(
        """
        from math import *
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Import(name="*", qualified_name="math", location=mock.ANY),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)

    _, stderr = capfd.readouterr()
    assert match_output(
        stderr,
        ["do not use 'from math import *' outside of __init__.py files, be explicit"],
    )


def test_root_context_import_from_submodule(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        from os.path import join
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Import(name="join", qualified_name="os.path.join", location=mock.ANY),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_import_multiple_starred(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        from os import *
        from math import *
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Import(name="*", qualified_name="os", location=mock.ANY),
            Import(name="*", qualified_name="math", location=mock.ANY),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_import_inside_stmt_block(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        if something:
            from math import *
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Import(name="*", qualified_name="math", location=mock.ANY),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_import_inside_function_block(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        def fn():
            from math import *
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Func(name="fn", location=mock.ANY, interface=CallInterface()),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_assignment(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        a = 42
        a = "reassigned!"

        b = hhgttg()

        x, y, = 1, 2
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Name("a", location=mock.ANY),
            Name("b", location=mock.ANY),
            Name("x", location=mock.ANY),
            Name("y", location=mock.ANY),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_ann_assignment(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        a: int = 42
        b: str = "an string!"
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Name("a", location=mock.ANY),
            Name("b", location=mock.ANY),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_aug_assignment(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        a = 7
        a += 5

        b %= "fmt"
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Name("a", location=mock.ANY),
            Name("b", location=mock.ANY),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_walrus(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        a = (b := 5)
        x = (y, z := 10)  # this is not `(y, z) := 10` but `y, (z := 10)`
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Name("a", location=mock.ANY),
            Name("b", location=mock.ANY),
            Name("x", location=mock.ANY),
            Name("z", location=mock.ANY),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_walrus_lambda(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        outer_lambda = (inner_lambda := lambda *a, **k: 0)
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Func(
                name="inner_lambda",
                location=mock.ANY,
                interface=CallInterface(vararg="a", kwarg="k"),
            ),
            Func(
                name="outer_lambda",
                location=mock.ANY,
                interface=CallInterface(vararg="a", kwarg="k"),
            ),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_faux_walrus_lambda(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        outer_lambda = (a, inner_lambda := lambda *a, **k: 0)
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Func(
                name="inner_lambda",
                location=mock.ANY,
                interface=CallInterface(vararg="a", kwarg="k"),
            ),
            Name(name="outer_lambda", location=mock.ANY),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_function_def(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        def fn_one(a, b):
            pass

        def fn_two(c):
            var = "i shouldn't be in the root context!"
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Func(
                name="fn_one",
                location=mock.ANY,
                interface=CallInterface(args=("a", "b")),
            ),
            Func(
                name="fn_two",
                location=mock.ANY,
                interface=CallInterface(args=("c",)),
            ),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_async_function_def(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        async def fn_one(a, b):
            pass

        async def fn_two(c):
            var = "i shouldn't be in the root context!"
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Func(
                name="fn_one",
                location=mock.ANY,
                interface=CallInterface(args=("a", "b")),
                is_async=True,
            ),
            Func(
                name="fn_two",
                location=mock.ANY,
                interface=CallInterface(args=("c",)),
                is_async=True,
            ),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_anonymous_lambda(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
    capfd: pytest.CaptureFixture[str],
):
    module_ast = parse(
        """
        lambda: 1
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)

    _, stderr = capfd.readouterr()
    assert match_output(stderr, ["top-level lambdas must be named"])


def test_root_context_lambda_named(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        x = lambda a, b: 1
        y: type = lambda: 1
        z = lambda *args: 1
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Func(
                "x",
                location=mock.ANY,
                interface=CallInterface(
                    args=(
                        "a",
                        "b",
                    )
                ),
            ),
            Func("y", location=mock.ANY, interface=CallInterface()),
            Func("z", location=mock.ANY, interface=CallInterface(vararg="args")),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_lambda_named_invalid_multiple_lhs_and_rhs(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
    capfd: pytest.CaptureFixture[str],
):
    module_ast = parse(
        """
        x, y = lambda a, b: 1, 2
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)

    _, stderr = capfd.readouterr()
    assert match_output(stderr, ["lambda assignment must be one-to-one"])


def test_root_context_lambda_named_invalid_multiple_lhs(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
    capfd: pytest.CaptureFixture[str],
):
    module_ast = parse(
        """
        x, y = lambda a: 1
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)

    _, stderr = capfd.readouterr()
    assert match_output(stderr, ["lambda assignment must be one-to-one"])


def test_root_context_lambda_named_invalid_multiple_rhs(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
    capfd: pytest.CaptureFixture[str],
):
    module_ast = parse(
        """
        x = lambda a, b: 1, 2
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)

    _, stderr = capfd.readouterr()
    assert match_output(stderr, ["lambda assignment must be one-to-one"])


def test_root_context_if(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        if some_variable == "some value":
            def fn(a):
                pass

            global_var = 40
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Func(name="fn", location=mock.ANY, interface=CallInterface(args=("a",))),
            Name(name="global_var", location=mock.ANY),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_nested_if(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        if some_variable == "some value":
            if some_other_variable == "some other value":
                def fn(a):
                    pass

                global_var = 40
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Func(name="fn", location=mock.ANY, interface=CallInterface(args=("a",))),
            Name(name="global_var", location=mock.ANY),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_for(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        for i in ITER:
            def fn(a):
                pass

            global_var = 40
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Func(name="fn", location=mock.ANY, interface=CallInterface(args=("a",))),
            Name(name="global_var", location=mock.ANY),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_nested_for(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        for iter in ITER_OF_ITERS:
            for i in iter:
                def fn(a):
                    pass

                global_var = 40
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Func(name="fn", location=mock.ANY, interface=CallInterface(args=("a",))),
            Name(name="global_var", location=mock.ANY),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_async_for(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        async for i in ITER:
            def fn(a):
                pass

            global_var = 40
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Func(name="fn", location=mock.ANY, interface=CallInterface(args=("a",))),
            Name(name="global_var", location=mock.ANY),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_while(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        while True:
            def fn(a):
                pass

            global_var = 40
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Func(name="fn", location=mock.ANY, interface=CallInterface(args=("a",))),
            Name(name="global_var", location=mock.ANY),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_nested_while(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        while True:
            while True:
                def fn(a):
                    pass

                global_var = 40
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Func(name="fn", location=mock.ANY, interface=CallInterface(args=("a",))),
            Name(name="global_var", location=mock.ANY),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_try(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        try:
            global_a = 1
        except TypeError:
            global_b = 2
        except:
            global_c = 3
        else:
            global_d = 4
        finally:
            global_e = 5
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Name("global_a", location=mock.ANY),
            Name("global_b", location=mock.ANY),
            Name("global_c", location=mock.ANY),
            Name("global_d", location=mock.ANY),
            Name("global_e", location=mock.ANY),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_nested_try(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        try:
            try:
                global_a = 1
            except TypeError:
                global_b = 2
            except:
                global_c = 3
            else:
                global_d = 4
            finally:
                global_e = 5
        finally:
            global_z = 26
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Name("global_a", location=mock.ANY),
            Name("global_b", location=mock.ANY),
            Name("global_c", location=mock.ANY),
            Name("global_d", location=mock.ANY),
            Name("global_e", location=mock.ANY),
            Name("global_z", location=mock.ANY),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_with(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        with context_manager() as ctx:
            def fn(a):
                pass

            global_var = 40
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Func(name="fn", location=mock.ANY, interface=CallInterface(args=("a",))),
            Name(name="global_var", location=mock.ANY),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_nested_with(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        with context_manager() as ctx:
            with ctx() as inner_ctx:
                def fn(a):
                    pass

                global_var = 40
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Func(name="fn", location=mock.ANY, interface=CallInterface(args=("a",))),
            Name(name="global_var", location=mock.ANY),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_async_with(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        async with context_manager() as ctx:
            def fn(a):
                pass

            global_var = 40
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Func(name="fn", location=mock.ANY, interface=CallInterface(args=("a",))),
            Name(name="global_var", location=mock.ANY),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_class_def(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    module_ast = parse(
        """
        class TopLevel:
            def __init__(self, a):
                self.a = a

            def method(self):
                var = "i shouldn't be in the root context!"
                return var + " see, i'm not!"
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [
            Class(
                name="TopLevel",
                location=mock.ANY,
                interface=CallInterface(
                    args=(
                        "self",
                        "a",
                    )
                ),
            ),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)


def test_root_context_module_level_docstring(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
    capfd: pytest.CaptureFixture[str],
):
    module_ast = parse(
        """
        '''
        Docstring at module level.
        '''
        """
    )
    context = compile_root_context(module_ast)
    expected = make_symbol_table(
        [],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)

    _, stderr = capfd.readouterr()
    assert match_output(stderr, [])


def test_root_context_module(
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
    capfd: pytest.CaptureFixture[str],
):
    module_ast = parse(
        """
        '''My module.'''
        from __future__ import annotations

        import a_module

        from another_module import some_function

        def my_function(a, b):
            sum = a + b
            return sum * sum

        class MyClass:
            def __init__(self):
                self.data = list()

            def get_data(self):
                return self.data
        """
    )

    with mock.patch("sys.exit") as m_exit:
        context = compile_root_context(module_ast)

    expected = make_symbol_table(
        [
            Import(
                name="annotations",
                qualified_name="__future__.annotations",
                location=mock.ANY,
            ),
            Import(name="a_module", location=mock.ANY),
            Import(
                name="some_function",
                qualified_name="another_module.some_function",
                location=mock.ANY,
            ),
            Func(
                name="my_function",
                location=mock.ANY,
                interface=CallInterface(args=("a", "b")),
            ),
            Class(
                name="MyClass",
                location=mock.ANY,
                interface=CallInterface(args=("self",)),
            ),
        ],
        include_root_symbols=True,
    )

    assert context.parent is None
    assert compare_symbol_table_symbols(context.symbol_table, expected)

    assert m_exit.call_count == 2

    _, stderr = capfd.readouterr()
    assert match_output(
        stderr,
        [
            "unable to find module 'a_module'",
            "unable to find module 'another_module'",
        ],
    )



from __future__ import annotations

from functools import partial
from pathlib import Path
from unittest import mock

import pytest

from rattr import error


class TestError:
    def test_rattr(self, capfd):
        _message = "the error message"

        with mock.patch("sys.exit") as _exit:
            error.rattr(_message)

        _, stderr = capfd.readouterr()

        assert "rattr" in stderr
        assert _message in stderr
        assert not _exit.called

    def test_info(self, capfd, arguments):
        _message = "the error message"

        with arguments(_warning_level="all"):
            with mock.patch("sys.exit") as _exit:
                error.info(_message)

        _, stderr = capfd.readouterr()

        assert "info" in stderr
        assert _message in stderr
        assert not _exit.called

    def test_warning(self, capfd):
        _message = "the error message"

        with mock.patch("sys.exit") as _exit:
            error.warning(_message)

        _, stderr = capfd.readouterr()

        assert "warning" in stderr
        assert _message in stderr
        assert not _exit.called

    def test_error(self, capfd):
        _message = "the error message"

        with mock.patch("sys.exit") as _exit:
            error.error(_message)

        _, stderr = capfd.readouterr()

        assert "error" in stderr
        assert _message in stderr
        assert not _exit.called

    def test_fatal(self, capfd):
        _message = "the error message"

        with mock.patch("sys.exit") as _exit:
            error.fatal(_message)

        _, stderr = capfd.readouterr()

        assert "fatal" in stderr
        assert _message in stderr
        assert _exit.called and _exit.call_count == 1


class TestErrorBadness:
    @pytest.fixture
    def reset(self, state):
        return partial(
            state,
            badness_from_simplification=0,
            badness_from_target_file=0,
            badness_from_imports=0,
            current_file=Path("target.py"),
        )

    def test_rattr(self, config, reset):
        # `error.rattr` takes badness to have a consistent interface, but does not
        # affect badness
        with reset():
            assert config.state.badness == 0
            error.rattr("blah", badness=1)
            assert config.state.badness == 0

    def test_info(self, config, reset):
        with reset():
            assert config.state.badness == 0
            error.info("blah", badness=1)
            assert config.state.badness == 1

    def test_warning(self, config, reset):
        with reset():
            assert config.state.badness == 0
            error.warning("blah", badness=1)
            assert config.state.badness == 1

    def test_error(self, config, reset):
        with reset():
            assert config.state.badness == 0
            error.error("blah", badness=1)
            assert config.state.badness == 1

    def test_fatal(self, config, reset):
        with reset():
            assert config.state.badness == 0
            with pytest.raises(SystemExit):
                error.fatal("blah", badness=1)
            assert config.state.badness == 1



from __future__ import annotations

from rattr.models.symbol.util import (
    get_basename_from_name,
    with_call_brackets,
    without_call_brackets,
)


class WithCallBrackets:
    def test_the_empty_string(self):
        assert with_call_brackets("") == ""

    def test_no_call_brackets(self):
        assert (
            with_call_brackets("i_dont_have_call_brackets")
            == "i_dont_have_call_brackets()"
        )

    def test_call_brackets(self):
        assert (
            with_call_brackets("keep_my_call_brackets_please()")
            == "keep_my_call_brackets_please()"
        )

    def test_just_brackets(self):
        assert with_call_brackets("()") == "()"
        assert with_call_brackets(")(") == ")(()"

    def test_separated_brackets(self):
        assert (
            with_call_brackets("do_not_count(my_brackets)")
            == "do_not_count(my_brackets)()"
        )


class TestWithoutCallBrackets:
    def test_the_empty_string(self):
        assert without_call_brackets("") == ""

    def test_no_call_brackets(self):
        assert (
            without_call_brackets("i_dont_have_call_brackets")
            == "i_dont_have_call_brackets"
        )

    def test_call_brackets(self):
        assert (
            without_call_brackets("remove_my_call_brackets")
            == "remove_my_call_brackets"
        )

    def test_just_brackets(self):
        assert without_call_brackets("()") == ""
        assert without_call_brackets(")(") == ")("

    def test_separated_brackets(self):
        assert (
            without_call_brackets("do_not_remove(my_brackets)")
            == "do_not_remove(my_brackets)"
        )


class TestGetBasenameFromString:
    def test_the_empty_string(self):
        assert get_basename_from_name("") == ""

    def test_is_already_basename(self):
        assert get_basename_from_name("my_var") == "my_var"

    def test_is_dotted(self):
        assert get_basename_from_name("my_var.attr") == "my_var"
        assert get_basename_from_name("my_var.deeply.nested.attrs") == "my_var"

    def test_is_starred(self):
        assert get_basename_from_name("*my_var") == "my_var"

    def test_is_starred_and_dotted(self):
        assert get_basename_from_name("*my_var.attr") == "my_var"
        assert get_basename_from_name("*my_var.deeply.nested.attrs") == "my_var"
